# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any, Generic, Sequence, TypeVar

from . import egrpc
from .alignment import AxisAligned, AlignmentSignature, assert_privacy_axis, new_alignment_signature
from .accountants import Accountant
from .prisoner import Prisoner, SensitiveInt, SensitiveFloat
from .realexpr import RealExpr
from .util import integer, floating

T = TypeVar("T")

class PrivArrayBase(Generic[T], Prisoner[T], AxisAligned):
    _privacy_axis        : int
    _alignment_signature : AlignmentSignature

    def __init__(self,
                 value          : Any,
                 distance       : RealExpr,
                 privacy_axis   : int,
                 parents        : Sequence[PrivArrayBase[Any]],
                 accountant     : Accountant[Any] | None,
                 *,
                 keep_alignment : bool,
                 ) -> None:
        if privacy_axis < 0:
            raise ValueError("privacy_axis must be non-negative.")

        if len(parents) == 0 and keep_alignment:
            raise ValueError("keep_alignment=True requires at least one parent.")

        self._privacy_axis = privacy_axis

        if keep_alignment:
            assert_privacy_axis(*parents)
            self._alignment_signature = parents[0]._alignment_signature
        else:
            self._alignment_signature = new_alignment_signature()

        super().__init__(value=value, distance=distance, parents=parents, accountant=accountant)

    @egrpc.property
    def privacy_axis(self) -> int:
        return self._privacy_axis

    @egrpc.property
    def alignment_signature(self) -> int:
        return self._alignment_signature

    def renew_alignment_signature(self) -> None:
        self._alignment_signature = new_alignment_signature()

@egrpc.remoteclass
class SensitiveDimInt(SensitiveInt):
    _alignment_signature : AlignmentSignature
    _scale               : int

    def __init__(self,
                 value               : integer,
                 distance            : RealExpr,
                 alignment_signature : AlignmentSignature,
                 scale               : int                     = 1,
                 *,
                 parents             : Sequence[Prisoner[Any]] = [],
                 accountant          : Accountant[Any] | None  = None,
                 ) -> None:
        self._alignment_signature = alignment_signature
        self._scale = scale
        super().__init__(value, distance, parents=parents, accountant=accountant)

    @egrpc.property
    def alignment_signature(self) -> AlignmentSignature:
        return self._alignment_signature

    @egrpc.property
    def scale(self) -> int:
        return self._scale

    @egrpc.method
    def __neg__(self) -> SensitiveDimInt:
        return SensitiveDimInt(-self._value,
                               distance            = self._distance,
                               alignment_signature = self._alignment_signature,
                               scale               = -self._scale,
                               parents             = [self])

    @egrpc.multimethod
    def __add__(self, other: SensitiveDimInt) -> SensitiveDimInt | SensitiveInt:
        if self._alignment_signature == other._alignment_signature:
            return SensitiveDimInt(self._value + other._value,
                                   distance            = self._distance + other._distance,
                                   alignment_signature = self._alignment_signature,
                                   scale               = self._scale + other._scale,
                                   parents             = [self, other])
        return super().__add__(other)  # type: ignore[no-any-return]

    @__add__.register
    def _(self, other: integer) -> SensitiveInt:
        return super().__add__(other)  # type: ignore[no-any-return]

    @__add__.register
    def _(self, other: floating) -> SensitiveFloat:
        return super().__add__(other)  # type: ignore[no-any-return]

    @__add__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return super().__add__(other)  # type: ignore[no-any-return]

    @__add__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return super().__add__(other)  # type: ignore[no-any-return]

    @egrpc.multimethod
    def __sub__(self, other: SensitiveDimInt) -> SensitiveDimInt | SensitiveInt:
        if self._alignment_signature == other._alignment_signature:
            return SensitiveDimInt(self._value - other._value,
                                   distance            = self._distance + other._distance,
                                   alignment_signature = self._alignment_signature,
                                   scale               = self._scale - other._scale,
                                   parents             = [self, other])
        return super().__sub__(other)  # type: ignore[no-any-return]

    @__sub__.register
    def _(self, other: integer) -> SensitiveInt:
        return super().__sub__(other)  # type: ignore[no-any-return]

    @__sub__.register
    def _(self, other: floating) -> SensitiveFloat:
        return super().__sub__(other)  # type: ignore[no-any-return]

    @__sub__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return super().__sub__(other)  # type: ignore[no-any-return]

    @__sub__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return super().__sub__(other)  # type: ignore[no-any-return]

    @egrpc.multimethod
    def __mul__(self, other: integer) -> SensitiveDimInt:
        return SensitiveDimInt(self._value * other,
                               distance            = self._distance * abs(int(other)),
                               alignment_signature = self._alignment_signature,
                               scale               = self._scale * int(other),
                               parents             = [self])

    @__mul__.register
    def _(self, other: floating) -> SensitiveFloat:
        return super().__mul__(other)  # type: ignore[no-any-return]

    @egrpc.multimethod
    def __rmul__(self, other: integer) -> SensitiveDimInt:  # type: ignore[misc]
        return SensitiveDimInt(other * self._value,
                               distance            = self._distance * abs(int(other)),
                               alignment_signature = self._alignment_signature,
                               scale               = int(other) * self._scale,
                               parents             = [self])

    @__rmul__.register
    def _(self, other: floating) -> SensitiveFloat:
        return super().__rmul__(other)  # type: ignore[no-any-return]



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import numpy as np
import privjail as pj
import privjail.pandas as ppd

from bokeh.plotting import figure, show
from bokeh.models import LogColorMapper, ColorBar, LogTicker
from bokeh.palettes import Magma256

# LAT_MIN, LAT_MAX = 39.5, 40.5
# LON_MIN, LON_MAX = 116.0, 117.0
LAT_MIN, LAT_MAX = 39.7, 40.2
LON_MIN, LON_MAX = 116.2, 116.7

def main(args):
    df = ppd.read_csv("data/tdrive.csv", "schema/tdrive.json")

    print("Dataset loaded.")

    G = args.grid_size

    # binning for lat, lon
    lat_bins = np.arange(np.floor(LAT_MIN / G), np.floor(LAT_MAX / G) + 1) * G
    df["lat_bin"] = ppd.cut(df["latitude"], bins=list(lat_bins), labels=list(lat_bins[:-1]))

    lon_bins = np.arange(np.floor(LON_MIN / G), np.floor(LON_MAX / G) + 1) * G
    df["lon_bin"] = ppd.cut(df["longitude"], bins=list(lon_bins), labels=list(lon_bins[:-1]))

    df = df.dropna()

    print("Binning done.")

    # bound user contribution
    df = df.sample(frac=1).groupby("taxi_id").head(args.max_freq)

    print("User contribution bounded.")

    # groupby-count
    counts = df.groupby(["lat_bin", "lon_bin"]).size()
    print("Counts calculated.")

    # add noise
    noisy_counts = counts.reveal(eps=args.eps)
    print("Counts revealed.")

    visualize(noisy_counts)

    print()
    print("Consumed Privacy Budget:")
    print(pj.budgets_spent())

def visualize(ser):
    df = ser.reset_index(name="count")

    counts_2d = df.pivot(index="lat_bin", columns="lon_bin", values="count").values

    mapper = LogColorMapper(palette=Magma256, low=1, high=counts_2d.max())

    p = figure(
        title="Traffic Density Heatmap",
        x_axis_label="Longitude",
        y_axis_label="Latitude",
        x_range=(LON_MIN, LON_MAX),
        y_range=(LAT_MIN, LAT_MAX),
        frame_width=800,
        frame_height=800,
        toolbar_location="above",
    )

    p.image(
        image=[counts_2d],
        x=LON_MIN,
        y=LAT_MIN,
        dw=LON_MAX - LON_MIN,
        dh=LAT_MAX - LAT_MIN,
        color_mapper=mapper
    )

    color_bar = ColorBar(
        color_mapper=mapper,
        ticker=LogTicker(),
        label_standoff=12,
        border_line_color=None,
        location=(0, 0)
    )

    p.add_layout(color_bar, 'right')

    show(p)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--remote", action="store_true")
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=12345)
    parser.add_argument("--eps", type=float, default=2.0)
    parser.add_argument("--max-freq", type=int, default=4)
    parser.add_argument("--grid-size", type=float, default=0.004)
    args = parser.parse_args()

    if args.remote:
        pj.connect(args.host, args.port)

    main(args)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any, Sequence
import math

from .util import Accountant, ParallelAccountant, SubsamplingAccountant
from .approx import ApproxDPAccountant
from .. import egrpc
from ..util import DPError

# Budget type: {alpha: epsilon} mapping
RDPBudgetType = dict[float, float]

def rdp_to_approx_eps(rdp_budget: RDPBudgetType, delta: float) -> float:
    nonzero = [(a, eps) for a, eps in rdp_budget.items() if eps > 0]
    if not nonzero:
        return 0.0

    # Rényi Differential Privacy
    # https://arxiv.org/pdf/1702.07476
    # Proposition 3: (α, ε)-RDP implies (ε + log(1/δ)/(α-1), δ)-DP
    return min(eps + math.log(1 / delta) / (a - 1) for a, eps in nonzero)

@egrpc.remoteclass
class RDPAccountant(Accountant[RDPBudgetType]):
    _alpha : list[float]
    _delta : float

    def __init__(self,
                 *,
                 alpha        : Sequence[float] | None = None,
                 budget_limit : RDPBudgetType | None   = None,
                 parent       : Accountant[Any] | None = None,
                 delta        : float | None           = None,
                 prepaid      : bool                   = False,
                 ):
        if alpha is None:
            if isinstance(parent, (RDPAccountant, RDPParallelAccountant, RDPSubsamplingAccountant)):
                self._alpha = parent._alpha
            else:
                raise ValueError("alpha must be specified when parent is not RDPAccountant/RDPParallelAccountant/RDPSubsamplingAccountant")
        else:
            if len(alpha) == 0:
                raise ValueError("alpha must not be empty")
            for a in alpha:
                if a <= 1:
                    raise ValueError(f"alpha must be > 1, got {a}")
            if len(set(alpha)) != len(alpha):
                raise ValueError("alpha must be unique")
            self._alpha = list(alpha)

        if isinstance(parent, ApproxDPAccountant):
            if delta is None:
                raise ValueError("delta must be specified when parent is ApproxDPAccountant")
            if not (0 < delta < 1):
                raise ValueError("delta must be in (0, 1)")
            # Spend delta ahead of time (same pattern as zCDPAccountant)
            parent.spend((0, delta))
            self._delta = delta
        else:
            self._delta = 0.0  # Not used when parent is not ApproxDPAccountant

        super().__init__(budget_limit=budget_limit, parent=parent, prepaid=prepaid)

    @egrpc.property
    def alpha(self) -> list[float]:
        return self._alpha

    @staticmethod
    def family_name() -> str:
        return "RDP"

    @egrpc.property
    def budget_spent(self) -> RDPBudgetType:
        return self._budget_spent

    def propagate(self, next_budget_spent: RDPBudgetType, parent: Accountant[Any]) -> None:
        if isinstance(parent, RDPParallelAccountant):
            parent.spend(next_budget_spent)
        elif isinstance(parent, RDPSubsamplingAccountant):
            parent.spend(next_budget_spent)
        elif isinstance(parent, ApproxDPAccountant):
            prev_eps = rdp_to_approx_eps(self._budget_spent, self._delta)
            next_eps = rdp_to_approx_eps(next_budget_spent, self._delta)
            parent.spend((next_eps - prev_eps, 0))
        else:
            raise Exception

    def compose(self, budget1: RDPBudgetType, budget2: RDPBudgetType) -> RDPBudgetType:
        # Rényi Differential Privacy
        # https://arxiv.org/pdf/1702.07476
        # Proposition 1: Composition is additive for each α
        return {a: budget1[a] + budget2[a] for a in self._alpha}

    def zero(self) -> RDPBudgetType:
        return {a: 0.0 for a in self._alpha}

    def exceeds(self, budget1: RDPBudgetType, budget2: RDPBudgetType) -> bool:
        return any(budget1[a] > budget2[a] for a in self._alpha)

    def assert_budget(self, budget: RDPBudgetType) -> None:
        assert set(budget.keys()) == set(self._alpha)
        for eps in budget.values():
            assert eps >= 0

    @classmethod
    def normalize_budget(cls, budget: Any) -> RDPBudgetType | None:
        if budget is None:
            return None
        elif isinstance(budget, dict):
            normalized = {float(k): float(v) for k, v in budget.items()}
            for eps in normalized.values():
                assert eps >= 0
            return normalized
        else:
            raise TypeError("RDP accountant budget must be a dict[float, float].")

    @staticmethod
    def parallel_accountant() -> type[RDPParallelAccountant]:
        return RDPParallelAccountant

    @staticmethod
    def subsampling_accountant() -> type[SubsamplingAccountant[RDPBudgetType]]:
        return RDPSubsamplingAccountant

class RDPParallelAccountant(ParallelAccountant[RDPBudgetType]):
    _alpha: list[float]

    def __init__(self,
                 *,
                 alpha        : Sequence[float] | None = None,
                 budget_limit : RDPBudgetType | None   = None,
                 parent       : Accountant[Any] | None = None,
                 ):
        if alpha is not None:
            if len(set(alpha)) != len(alpha):
                raise ValueError("alpha must be unique")
            self._alpha = list(alpha)
        elif isinstance(parent, RDPAccountant):
            self._alpha = parent._alpha
        elif isinstance(parent, RDPParallelAccountant):
            self._alpha = parent._alpha
        else:
            raise ValueError("alpha must be specified or parent must be RDPAccountant/RDPParallelAccountant")

        super().__init__(budget_limit=budget_limit, parent=parent)

    @staticmethod
    def family_name() -> str:
        return "RDP"

    def propagate(self, next_budget_spent: RDPBudgetType, parent: Accountant[Any]) -> None:
        if isinstance(parent, RDPAccountant):
            # Parallel composition: propagate the difference
            # Both next_budget_spent and _budget_spent have all alpha
            diff = {a: next_budget_spent[a] - self._budget_spent[a] for a in self._alpha}
            parent.spend(diff)
        else:
            raise Exception

    def compose(self, budget1: RDPBudgetType, budget2: RDPBudgetType) -> RDPBudgetType:
        # Parallel composition: take max for each alpha
        return {a: max(budget1[a], budget2[a]) for a in self._alpha}

    def zero(self) -> RDPBudgetType:
        return {a: 0.0 for a in self._alpha}

    def exceeds(self, budget1: RDPBudgetType, budget2: RDPBudgetType) -> bool:
        return any(budget1[a] > budget2[a] for a in self._alpha)

    def assert_budget(self, budget: RDPBudgetType) -> None:
        assert set(budget.keys()) == set(self._alpha)
        for eps in budget.values():
            assert eps >= 0

    @classmethod
    def normalize_budget(cls, budget: Any) -> RDPBudgetType | None:
        return RDPAccountant.normalize_budget(budget)

class RDPSubsamplingAccountant(SubsamplingAccountant[RDPBudgetType]):
    """Subsampling accountant for RDP with Gaussian mechanism.

    This accountant enforces single-use: only one spend() call is allowed,
    because the subsampled Gaussian RDP formula assumes one mechanism
    application per subsampling operation.

    The budget passed to spend() should already be the subsampled RDP budget
    (computed by the mechanism using the Mironov formula).
    """
    _alpha : list[float]
    _used  : bool

    def __init__(self,
                 *,
                 alpha         : Sequence[float] | None = None,
                 sampling_rate : float,
                 budget_limit  : RDPBudgetType | None   = None,
                 parent        : Accountant[Any] | None = None,
                 ):
        if alpha is not None:
            if len(set(alpha)) != len(alpha):
                raise ValueError("alpha must be unique")
            self._alpha = list(alpha)
        elif isinstance(parent, RDPAccountant):
            self._alpha = parent._alpha
        elif isinstance(parent, RDPParallelAccountant):
            self._alpha = parent._alpha
        else:
            raise ValueError("alpha must be specified or parent must be RDPAccountant/RDPParallelAccountant")

        self._used = False
        super().__init__(sampling_rate=sampling_rate, budget_limit=budget_limit, parent=parent)

    @staticmethod
    def family_name() -> str:
        return "RDP"

    def spend(self, budget: RDPBudgetType) -> None:
        if self._used:
            raise DPError(
                "RDPSubsamplingAccountant can only be used once. "
                "Subsampled Gaussian RDP assumes one mechanism application per subsampling."
            )
        self._used = True
        super().spend(budget)

    def propagate(self, next_budget_spent: RDPBudgetType, parent: Accountant[Any]) -> None:
        if isinstance(parent, RDPAccountant):
            parent.spend(next_budget_spent)
        else:
            raise Exception

    def compose(self, budget1: RDPBudgetType, budget2: RDPBudgetType) -> RDPBudgetType:
        del budget1
        return budget2

    def zero(self) -> RDPBudgetType:
        return {a: 0.0 for a in self._alpha}

    def exceeds(self, budget1: RDPBudgetType, budget2: RDPBudgetType) -> bool:
        return any(budget1[a] > budget2[a] for a in self._alpha)

    def assert_budget(self, budget: RDPBudgetType) -> None:
        assert set(budget.keys()) == set(self._alpha)
        for eps in budget.values():
            assert eps >= 0

    @classmethod
    def normalize_budget(cls, budget: Any) -> RDPBudgetType | None:
        return RDPAccountant.normalize_budget(budget)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, Sequence, Any, overload
import math

import numpy as _np
import pandas as _pd
import numpy.typing as _npt

from .util import DPError, floating, realnum
from .accountants import BudgetType, PureDPAccountant, ApproxDPAccountant, zCDPAccountant, RDPAccountant, RDPBudgetType, RDPSubsamplingAccountant
from .prisoner import Prisoner, SensitiveInt, SensitiveFloat
from .numpy import SensitiveNDArray
from .pandas import SensitiveSeries, SensitiveDataFrame
from . import egrpc

T = TypeVar("T")

def assert_sensitivity(sensitivity: realnum) -> None:
    if sensitivity == math.inf:
        raise DPError("Unbounded sensitivity")

    if sensitivity <= 0:
        raise DPError(f"Invalid sensitivity ({sensitivity})")

def assert_eps(eps: floating) -> None:
    if eps <= 0:
        raise DPError(f"Invalid epsilon ({eps})")

def assert_delta(delta: floating) -> None:
    if delta <= 0 or 1.0 < delta:
        raise DPError(f"Invalid delta ({delta})")

def assert_rho(rho: floating) -> None:
    if rho <= 0:
        raise DPError(f"Invalid rho ({rho})")

def assert_laplace_scale(scale: floating) -> None:
    if scale <= 0:
        raise DPError(f"Invalid scale ({scale})")

def assert_gaussian_scale(scale: floating) -> None:
    if scale <= 0:
        raise DPError(f"Invalid scale ({scale})")

def resolve_laplace_params(sensitivity : float,
                           *,
                           eps         : float | None = None,
                           scale       : float | None = None,
                           ) -> tuple[float, float]:
    if eps is None and scale is None:
        raise ValueError("Either eps or scale must be specified.")

    elif eps is not None and scale is not None:
        raise ValueError("eps and scale cannot be specified simultaneously.")

    elif eps is not None:
        assert_eps(eps)
        resolved_scale = sensitivity / eps
        assert_laplace_scale(resolved_scale)
        return eps, resolved_scale

    elif scale is not None:
        assert_laplace_scale(scale)
        resolved_eps = sensitivity / scale
        assert_eps(resolved_eps)
        return resolved_eps, scale

    else:
        raise Exception

def sigma_from_eps_delta(sensitivity: float, eps: float, delta: float) -> float:
    # The Algorithmic Foundations of Differential Privacy
    # https://www.cis.upenn.edu/~aaroth/Papers/privacybook.pdf
    # Theorem 3.22
    return sensitivity * math.sqrt(2.0 * math.log(1.25 / delta)) / eps

def eps_from_sigma_delta(sensitivity: float, sigma: float, delta: float) -> float:
    return sensitivity * math.sqrt(2.0 * math.log(1.25 / delta)) / sigma

def sigma_from_rho(sensitivity: float, rho: float) -> float:
    # Concentrated Differential Privacy: Simplifications, Extensions, and Lower Bounds
    # https://arxiv.org/pdf/1605.02065
    # Proposition 1.6
    return sensitivity / math.sqrt(2.0 * rho)

def rho_from_sigma(sensitivity: float, sigma: float) -> float:
    return sensitivity * sensitivity / (2.0 * sigma * sigma)

def rdp_eps_from_sigma(sensitivity: float, sigma: float, alpha: float) -> float:
    # Rényi Differential Privacy
    # https://arxiv.org/pdf/1702.07476
    # Example 1: Gaussian mechanism satisfies (α, αΔ²/(2σ²))-RDP
    return alpha * sensitivity * sensitivity / (2.0 * sigma * sigma)

def subsampled_rdp_eps_from_sigma(sensitivity   : float,
                                  sigma         : float,
                                  alpha         : float,
                                  sampling_rate : float,
                                  ) -> float:
    if sampling_rate <= 0 or sampling_rate > 1:
        raise ValueError("sampling_rate must be in (0, 1]")

    if not float(alpha).is_integer() or alpha < 2:
        raise ValueError("alpha must be an integer >= 2 for subsampled RDP")

    if sampling_rate == 1.0:
        return rdp_eps_from_sigma(sensitivity, sigma, alpha)

    # Rényi Differential Privacy of the Sampled Gaussian Mechanism
    # https://arxiv.org/abs/1908.10530
    # Theorem 9: A_α = Σ_{k=0}^{α} C(α,k) (1-q)^{α-k} q^k exp((k²-k) Δ²/(2σ²))
    # ε(α) = log(A_α) / (α-1)
    alpha_int = int(alpha)
    q = sampling_rate
    c = (sensitivity ** 2) / (2.0 * sigma ** 2)

    k = _np.arange(alpha_int + 1, dtype=float)
    log_binom = _np.array([math.lgamma(alpha_int + 1) - math.lgamma(ki + 1) - math.lgamma(alpha_int - ki + 1) for ki in k])
    log_terms = log_binom + (alpha_int - k) * _np.log(1 - q) + k * _np.log(q) + (k * k - k) * c
    log_a = float(_np.logaddexp.reduce(log_terms))

    return log_a / (alpha - 1)

def resolve_gaussian_params_approx(sensitivity : float,
                                   *,
                                   eps         : float | None = None,
                                   delta       : float | None = None,
                                   scale       : float | None = None,
                                   ) -> tuple[float, float]:
    if delta is None:
        raise ValueError("delta must be specified when using the Gaussian mechanism under approx DP.")

    elif eps is None and scale is None:
        raise ValueError("Either eps or scale must be specified.")

    elif eps is not None and scale is not None:
        raise ValueError("eps and scale cannot be specified simultaneously.")

    elif eps is not None:
        assert_eps(eps)
        resolved_scale = sigma_from_eps_delta(sensitivity, eps, delta)
        assert_gaussian_scale(resolved_scale)
        return eps, resolved_scale

    elif scale is not None:
        assert_gaussian_scale(scale)
        resolved_eps = eps_from_sigma_delta(sensitivity, scale, delta)
        assert_eps(resolved_eps)
        return resolved_eps, scale

    else:
        raise Exception

def resolve_gaussian_params_zcdp(sensitivity : float,
                                 *,
                                 rho         : float | None = None,
                                 scale       : float | None = None,
                                 ) -> tuple[float, float]:
    if rho is None and scale is None:
        raise ValueError("Either rho or scale must be specified.")

    elif rho is not None and scale is not None:
        raise ValueError("rho and scale cannot be specified simultaneously.")

    elif rho is not None:
        assert_rho(rho)
        resolved_scale = sigma_from_rho(sensitivity, rho)
        assert_gaussian_scale(resolved_scale)
        return rho, resolved_scale

    elif scale is not None:
        assert_gaussian_scale(scale)
        resolved_rho = rho_from_sigma(sensitivity, scale)
        assert_rho(resolved_rho)
        return resolved_rho, scale

    else:
        raise Exception

def resolve_gaussian_params_rdp(sensitivity   : float,
                                alpha         : list[float],
                                *,
                                scale         : float | None = None,
                                sampling_rate : float | None = None,
                                ) -> tuple[RDPBudgetType, float]:
    if scale is None:
        raise ValueError("scale must be specified when using the Gaussian mechanism under RDP.")

    assert_gaussian_scale(scale)

    if sampling_rate is None:
        budget = {a: rdp_eps_from_sigma(sensitivity, scale, a) for a in alpha}
    else:
        budget = {a: subsampled_rdp_eps_from_sigma(sensitivity, scale, a, sampling_rate) for a in alpha}

    return budget, scale

@overload
def laplace_mechanism(prisoner : SensitiveInt | SensitiveFloat,
                      *,
                      eps      : float | None = ...,
                      scale    : float | None = ...,
                      ) -> float: ...

@overload
def laplace_mechanism(prisoner : SensitiveSeries[Any],
                      *,
                      eps      : float | None = ...,
                      scale    : float | None = ...,
                      ) -> _pd.Series: ... # type: ignore[type-arg]

@overload
def laplace_mechanism(prisoner : SensitiveDataFrame,
                      *,
                      eps      : float | None = ...,
                      scale    : float | None = ...,
                      ) -> _pd.DataFrame: ...

@overload
def laplace_mechanism(prisoner : SensitiveNDArray,
                      *,
                      eps      : float | None = ...,
                      scale    : float | None = ...,
                      ) -> _npt.NDArray[Any]: ...

@egrpc.multifunction
def laplace_mechanism(prisoner : SensitiveInt | SensitiveFloat,
                      *,
                      eps      : float | None = None,
                      scale    : float | None = None,
                      ) -> float:
    sensitivity = float(prisoner._distance.max())
    assert_sensitivity(sensitivity)

    resolved_eps, resolved_scale = resolve_laplace_params(sensitivity, eps=eps, scale=scale)

    result = float(_np.random.laplace(loc=prisoner._value, scale=resolved_scale))

    if isinstance(prisoner.accountant, PureDPAccountant):
        prisoner.accountant.spend(resolved_eps)
    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        prisoner.accountant.spend((resolved_eps, 0.0))
    else:
        raise RuntimeError

    return result

@laplace_mechanism.register # type: ignore
def _(prisoner : SensitiveSeries[realnum],
      *,
      eps      : float | None = None,
      scale    : float | None = None,
      ) -> _pd.Series: # type: ignore[type-arg]
    if prisoner._distance_group_axes == (0,):
        assert isinstance(prisoner._partitioned_distances, list)

        if eps is not None:
            assert_eps(eps)
            eps_each = eps / len(prisoner)
            scales = []
            for distance in prisoner._partitioned_distances:
                sensitivity = float(distance.max())
                assert_sensitivity(sensitivity)
                _, resolved_scale = resolve_laplace_params(sensitivity, eps=eps_each)
                scales.append(resolved_scale)
            data = _np.random.laplace(loc=prisoner._value, scale=scales)
            spent_eps = eps

        elif scale is not None:
            assert_laplace_scale(scale)
            total_eps = 0.0
            for distance in prisoner._partitioned_distances:
                sensitivity = float(distance.max())
                assert_sensitivity(sensitivity)
                resolved_eps, _ = resolve_laplace_params(sensitivity, scale=scale)
                total_eps += resolved_eps
            data = _np.random.laplace(loc=prisoner._value, scale=scale)
            spent_eps = total_eps

        else:
            raise Exception

    else:
        sensitivity = float(prisoner._distance.max())
        resolved_eps, resolved_scale = resolve_laplace_params(sensitivity, eps=eps, scale=scale)
        data = _np.random.laplace(loc=prisoner._value, scale=resolved_scale)
        spent_eps = resolved_eps

    if isinstance(prisoner.accountant, PureDPAccountant):
        prisoner.accountant.spend(spent_eps)
    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        prisoner.accountant.spend((spent_eps, 0.0))
    else:
        raise RuntimeError

    return _pd.Series(data, index=prisoner.index, name=prisoner.name)

@laplace_mechanism.register # type: ignore
def _(prisoner : SensitiveDataFrame,
      *,
      eps      : float | None = None,
      scale    : float | None = None,
      ) -> _pd.DataFrame:
    if prisoner._distance_group_axes == (1,):
        assert isinstance(prisoner._partitioned_distances, list)

        if eps is not None:
            assert_eps(eps)
            eps_each = eps / len(prisoner)
            scales = []
            for distance in prisoner._partitioned_distances:
                sensitivity = float(distance.max())
                assert_sensitivity(sensitivity)
                _, resolved_scale = resolve_laplace_params(sensitivity, eps=eps_each)
                scales.append(resolved_scale)
            data = _np.random.laplace(loc=prisoner._value, scale=scales)
            spent_eps = eps

        elif scale is not None:
            assert_laplace_scale(scale)
            total_eps = 0.0
            for distance in prisoner._partitioned_distances:
                sensitivity = float(distance.max())
                assert_sensitivity(sensitivity)
                resolved_eps, _ = resolve_laplace_params(sensitivity, scale=scale)
                total_eps += resolved_eps
            data = _np.random.laplace(loc=prisoner._value, scale=scale)
            spent_eps = total_eps

        else:
            raise Exception

    else:
        sensitivity = float(prisoner._distance.max())
        resolved_eps, resolved_scale = resolve_laplace_params(sensitivity, eps=eps, scale=scale)
        data = _np.random.laplace(loc=prisoner._value, scale=resolved_scale)
        spent_eps = resolved_eps

    if isinstance(prisoner.accountant, PureDPAccountant):
        prisoner.accountant.spend(spent_eps)
    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        prisoner.accountant.spend((spent_eps, 0.0))
    else:
        raise RuntimeError

    return _pd.DataFrame(data, index=prisoner.index, columns=prisoner.columns)

@laplace_mechanism.register # type: ignore
def _(prisoner : SensitiveNDArray,
      *,
      eps      : float | None = None,
      scale    : float | None = None,
      ) -> _npt.NDArray[Any]:
    sensitivity = float(prisoner._distance.max())
    assert_sensitivity(sensitivity)

    resolved_eps, resolved_scale = resolve_laplace_params(sensitivity, eps=eps, scale=scale)
    samples = _np.random.laplace(loc=prisoner._value, scale=resolved_scale)

    if isinstance(prisoner.accountant, PureDPAccountant):
        prisoner.accountant.spend(resolved_eps)
    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        prisoner.accountant.spend((resolved_eps, 0.0))
    else:
        raise RuntimeError

    return _np.asarray(samples)

@overload
def gaussian_mechanism(prisoner : SensitiveInt | SensitiveFloat,
                       *,
                       eps      : float | None = ...,
                       delta    : float | None = ...,
                       rho      : float | None = ...,
                       scale    : float | None = ...,
                       ) -> float: ...

@overload
def gaussian_mechanism(prisoner : SensitiveSeries[Any],
                       *,
                       eps      : float | None = ...,
                       delta    : float | None = ...,
                       rho      : float | None = ...,
                       scale    : float | None = ...,
                       ) -> _pd.Series: ... # type: ignore[type-arg]

@overload
def gaussian_mechanism(prisoner : SensitiveDataFrame,
                       *,
                       eps      : float | None = ...,
                       delta    : float | None = ...,
                       rho      : float | None = ...,
                       scale    : float | None = ...,
                       ) -> _pd.DataFrame: ...

@overload
def gaussian_mechanism(prisoner : SensitiveNDArray,
                       *,
                       eps      : float | None = ...,
                       delta    : float | None = ...,
                       rho      : float | None = ...,
                       scale    : float | None = ...,
                       ) -> _npt.NDArray[Any]: ...

@egrpc.multifunction
def gaussian_mechanism(prisoner : SensitiveInt | SensitiveFloat,
                       *,
                       eps      : float | None = None,
                       delta    : float | None = None,
                       rho      : float | None = None,
                       scale    : float | None = None,
                       ) -> float:
    sensitivity = float(prisoner._distance.max())
    assert_sensitivity(sensitivity)

    budget : BudgetType

    if isinstance(prisoner.accountant, PureDPAccountant):
        raise DPError("Gaussian mechanism cannot be used under Pure DP")

    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        resolved_eps, resolved_scale = resolve_gaussian_params_approx(sensitivity, eps=eps, delta=delta, scale=scale)
        assert delta is not None
        budget = (resolved_eps, delta)

    elif isinstance(prisoner.accountant, zCDPAccountant):
        resolved_rho, resolved_scale = resolve_gaussian_params_zcdp(sensitivity, rho=rho, scale=scale)
        budget = resolved_rho

    elif isinstance(prisoner.accountant, RDPAccountant):
        # Check if under subsampling context
        sampling_rate: float | None = None
        if isinstance(prisoner.accountant.parent, RDPSubsamplingAccountant):
            sampling_rate = prisoner.accountant.parent.sampling_rate

        rdp_budget, resolved_scale = resolve_gaussian_params_rdp(
            sensitivity, prisoner.accountant.alpha, scale=scale, sampling_rate=sampling_rate
        )
        budget = rdp_budget

    else:
        raise RuntimeError

    result = float(_np.random.normal(loc=prisoner._value, scale=resolved_scale))

    prisoner.accountant.spend(budget)

    return result

@gaussian_mechanism.register # type: ignore
def _(prisoner : SensitiveSeries[realnum],
      *,
      eps      : float | None = None,
      delta    : float | None = None,
      rho      : float | None = None,
      scale    : float | None = None,
      ) -> _pd.Series: # type: ignore[type-arg]
    budget : BudgetType

    if isinstance(prisoner.accountant, PureDPAccountant):
        raise DPError("Gaussian mechanism cannot be used under Pure DP")

    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        assert delta is not None

        if prisoner._distance_group_axes == (0,):
            assert isinstance(prisoner._partitioned_distances, list)
            delta_each = delta / len(prisoner)

            if eps is not None:
                assert_eps(eps)
                eps_each = eps / len(prisoner)
                scales = []
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    _, resolved_scale = resolve_gaussian_params_approx(sensitivity, eps=eps_each, delta=delta_each)
                    scales.append(resolved_scale)
                data = _np.random.normal(loc=prisoner._value, scale=scales)
                spent_eps = eps

            elif scale is not None:
                assert_gaussian_scale(scale)
                total_eps = 0.0
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    resolved_eps, _ = resolve_gaussian_params_approx(sensitivity, delta=delta_each, scale=scale)
                    total_eps += resolved_eps
                data = _np.random.normal(loc=prisoner._value, scale=scale)
                spent_eps = total_eps

            else:
                raise Exception

            budget = (spent_eps, delta)

        else:
            sensitivity = float(prisoner._distance.max())
            resolved_eps, resolved_scale = resolve_gaussian_params_approx(sensitivity, eps=eps, delta=delta, scale=scale)
            data = _np.random.normal(loc=prisoner._value, scale=resolved_scale)
            budget = (resolved_eps, delta)

    elif isinstance(prisoner.accountant, zCDPAccountant):
        if prisoner._distance_group_axes == (0,):
            assert isinstance(prisoner._partitioned_distances, list)

            if rho is not None:
                assert_rho(rho)
                rho_each = rho / len(prisoner)
                scales = []
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    _, resolved_scale = resolve_gaussian_params_zcdp(sensitivity, rho=rho_each)
                    scales.append(resolved_scale)
                data = _np.random.normal(loc=prisoner._value, scale=scales)
                spent_rho = rho
            else:
                assert scale is not None
                assert_gaussian_scale(scale)
                total_rho = 0.0
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    resolved_rho, _ = resolve_gaussian_params_zcdp(sensitivity, scale=scale)
                    total_rho += resolved_rho
                data = _np.random.normal(loc=prisoner._value, scale=scale)
                spent_rho = total_rho

            budget = spent_rho

        else:
            sensitivity = float(prisoner._distance.max())
            resolved_rho, resolved_scale = resolve_gaussian_params_zcdp(sensitivity, rho=rho, scale=scale)
            data = _np.random.normal(loc=prisoner._value, scale=resolved_scale)
            budget = resolved_rho

    elif isinstance(prisoner.accountant, RDPAccountant):
        alpha = prisoner.accountant.alpha

        # Check if under subsampling context
        sampling_rate: float | None = None
        if isinstance(prisoner.accountant.parent, RDPSubsamplingAccountant):
            sampling_rate = prisoner.accountant.parent.sampling_rate

        if prisoner._distance_group_axes == (0,):
            assert isinstance(prisoner._partitioned_distances, list)

            # Subsampling not supported with partitioned distances
            if sampling_rate is not None:
                raise DPError("Subsampled RDP is not supported with partitioned distances")

            assert scale is not None
            assert_gaussian_scale(scale)
            total_budget: RDPBudgetType = {a: 0.0 for a in alpha}
            for distance in prisoner._partitioned_distances:
                sensitivity = float(distance.max())
                assert_sensitivity(sensitivity)
                elem_budget, _ = resolve_gaussian_params_rdp(sensitivity, alpha, scale=scale)
                for a in alpha:
                    total_budget[a] += elem_budget[a]
            data = _np.random.normal(loc=prisoner._value, scale=scale)
            budget = total_budget

        else:
            sensitivity = float(prisoner._distance.max())
            rdp_budget, resolved_scale = resolve_gaussian_params_rdp(
                sensitivity, alpha, scale=scale, sampling_rate=sampling_rate
            )
            data = _np.random.normal(loc=prisoner._value, scale=resolved_scale)
            budget = rdp_budget

    else:
        raise RuntimeError

    prisoner.accountant.spend(budget)

    return _pd.Series(data, index=prisoner.index, name=prisoner.name)

@gaussian_mechanism.register # type: ignore
def _(prisoner : SensitiveDataFrame,
      *,
      eps      : float | None = None,
      delta    : float | None = None,
      rho      : float | None = None,
      scale    : float | None = None,
      ) -> _pd.DataFrame:
    budget : BudgetType

    if isinstance(prisoner.accountant, PureDPAccountant):
        raise DPError("Gaussian mechanism cannot be used under Pure DP")

    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        assert delta is not None

        if prisoner._distance_group_axes == (1,):
            assert isinstance(prisoner._partitioned_distances, list)
            delta_each = delta / len(prisoner)

            if eps is not None:
                assert_eps(eps)
                eps_each = eps / len(prisoner)
                scales = []
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    _, resolved_scale = resolve_gaussian_params_approx(sensitivity, eps=eps_each, delta=delta_each)
                    scales.append(resolved_scale)
                data = _np.random.normal(loc=prisoner._value, scale=scales)
                spent_eps = eps

            elif scale is not None:
                assert_gaussian_scale(scale)
                total_eps = 0.0
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    resolved_eps, _ = resolve_gaussian_params_approx(sensitivity, delta=delta_each, scale=scale)
                    total_eps += resolved_eps
                data = _np.random.normal(loc=prisoner._value, scale=scale)
                spent_eps = total_eps

            else:
                raise Exception

            budget = (spent_eps, delta)

        else:
            sensitivity = float(prisoner._distance.max())
            resolved_eps, resolved_scale = resolve_gaussian_params_approx(sensitivity, eps=eps, delta=delta, scale=scale)
            data = _np.random.normal(loc=prisoner._value, scale=resolved_scale)
            budget = (resolved_eps, delta)

    elif isinstance(prisoner.accountant, zCDPAccountant):
        if prisoner._distance_group_axes == (1,):
            assert isinstance(prisoner._partitioned_distances, list)

            if rho is not None:
                assert_rho(rho)
                rho_each = rho / len(prisoner)
                scales = []
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    _, resolved_scale = resolve_gaussian_params_zcdp(sensitivity, rho=rho_each)
                    scales.append(resolved_scale)
                data = _np.random.normal(loc=prisoner._value, scale=scales)
                spent_rho = rho

            elif scale is not None:
                assert_gaussian_scale(scale)
                total_rho = 0.0
                for distance in prisoner._partitioned_distances:
                    sensitivity = float(distance.max())
                    assert_sensitivity(sensitivity)
                    resolved_rho, _ = resolve_gaussian_params_zcdp(sensitivity, scale=scale)
                    total_rho += resolved_rho
                data = _np.random.normal(loc=prisoner._value, scale=scale)
                spent_rho = total_rho

            else:
                raise Exception

            budget = spent_rho

        else:
            sensitivity = float(prisoner._distance.max())
            resolved_rho, resolved_scale = resolve_gaussian_params_zcdp(sensitivity, rho=rho, scale=scale)
            data = _np.random.normal(loc=prisoner._value, scale=resolved_scale)
            budget = resolved_rho

    elif isinstance(prisoner.accountant, RDPAccountant):
        alpha = prisoner.accountant.alpha

        # Check if under subsampling context
        sampling_rate: float | None = None
        if isinstance(prisoner.accountant.parent, RDPSubsamplingAccountant):
            sampling_rate = prisoner.accountant.parent.sampling_rate

        if prisoner._distance_group_axes == (1,):
            assert isinstance(prisoner._partitioned_distances, list)

            # Subsampling not supported with partitioned distances
            if sampling_rate is not None:
                raise DPError("Subsampled RDP is not supported with partitioned distances")

            assert scale is not None
            assert_gaussian_scale(scale)
            total_budget: RDPBudgetType = {a: 0.0 for a in alpha}
            for distance in prisoner._partitioned_distances:
                sensitivity = float(distance.max())
                assert_sensitivity(sensitivity)
                elem_budget, _ = resolve_gaussian_params_rdp(sensitivity, alpha, scale=scale)
                for a in alpha:
                    total_budget[a] += elem_budget[a]
            data = _np.random.normal(loc=prisoner._value, scale=scale)
            budget = total_budget

        else:
            sensitivity = float(prisoner._distance.max())
            rdp_budget, resolved_scale = resolve_gaussian_params_rdp(
                sensitivity, alpha, scale=scale, sampling_rate=sampling_rate
            )
            data = _np.random.normal(loc=prisoner._value, scale=resolved_scale)
            budget = rdp_budget

    else:
        raise RuntimeError

    prisoner.accountant.spend(budget)

    return _pd.DataFrame(data, index=prisoner.index, columns=prisoner.columns)

@gaussian_mechanism.register # type: ignore
def _(prisoner : SensitiveNDArray,
      *,
      eps      : float | None = None,
      delta    : float | None = None,
      rho      : float | None = None,
      scale    : float | None = None,
      ) -> _npt.NDArray[Any]:
    sensitivity = float(prisoner._distance.max())
    assert_sensitivity(sensitivity)

    budget: BudgetType

    if isinstance(prisoner.accountant, PureDPAccountant):
        raise DPError("Gaussian mechanism cannot be used under Pure DP")

    elif isinstance(prisoner.accountant, ApproxDPAccountant):
        resolved_eps, resolved_scale = resolve_gaussian_params_approx(sensitivity, eps=eps, delta=delta, scale=scale)
        assert delta is not None
        budget = (resolved_eps, delta)

    elif isinstance(prisoner.accountant, zCDPAccountant):
        resolved_rho, resolved_scale = resolve_gaussian_params_zcdp(sensitivity, rho=rho, scale=scale)
        budget = resolved_rho

    elif isinstance(prisoner.accountant, RDPAccountant):
        # Check if under subsampling context
        sampling_rate: float | None = None
        if isinstance(prisoner.accountant.parent, RDPSubsamplingAccountant):
            sampling_rate = prisoner.accountant.parent.sampling_rate

        rdp_budget, resolved_scale = resolve_gaussian_params_rdp(
            sensitivity, prisoner.accountant.alpha, scale=scale, sampling_rate=sampling_rate
        )
        budget = rdp_budget

    else:
        raise RuntimeError

    samples = _np.random.normal(loc=prisoner._value, scale=resolved_scale)

    prisoner.accountant.spend(budget)

    return _np.asarray(samples)

@egrpc.function
def exponential_mechanism(scores : Sequence[SensitiveInt | SensitiveFloat],
                          *,
                          eps    : float | None = None,
                          rho    : float | None = None,
                          ) -> int:
    if len(scores) == 0:
        raise ValueError("scores must have at least one element.")

    sensitivity = max([v._distance.max() for v in scores]) # type: ignore[type-var]
    assert_sensitivity(sensitivity)

    # create a dummy prisoner to propagate budget consumption to all prisoners
    prisoner_dummy = Prisoner(0, scores[0]._distance, parents=scores)

    budget : BudgetType

    if isinstance(prisoner_dummy.accountant, PureDPAccountant):
        assert eps is not None
        assert_eps(eps)
        budget = eps

    elif isinstance(prisoner_dummy.accountant, ApproxDPAccountant):
        assert eps is not None
        assert_eps(eps)
        budget = (eps, 0.0)

    elif isinstance(prisoner_dummy.accountant, zCDPAccountant):
        # Bounding, Concentrating, and Truncating: Unifying Privacy Loss Composition for Data Analytics
        # https://arxiv.org/pdf/2004.07223
        if eps is not None and rho is not None:
            raise ValueError("only one of eps and rho should be specified")
        elif eps is not None:
            assert_eps(eps)
            rho = eps * eps / 8.0
        elif rho is not None:
            assert_rho(rho)
            eps = math.sqrt(8.0 * rho)
        else:
            raise ValueError("eps or rho should be specified")

        budget = rho

    else:
        raise RuntimeError

    exponents = [eps * s._value / sensitivity / 2 for s in scores]

    # to prevent too small or large values (-> 0 or inf)
    M: float = _np.max(exponents) # type: ignore[arg-type]
    p = [_np.exp(x - M) for x in exponents]
    p /= sum(p)
    result = _np.random.choice(len(scores), p=p)

    prisoner_dummy.accountant.spend(budget)

    return result

def argmax(args : Sequence[SensitiveInt | SensitiveFloat],
           *,
           eps  : floating | None = None,
           rho  : floating | None = None,
           mech : str             = "exp",
           ) -> int:
    if mech == "exp":
        return exponential_mechanism(args,
                                     eps = float(eps) if eps is not None else None,
                                     rho = float(rho) if rho is not None else None)
    else:
        raise ValueError(f"Unknown DP mechanism: '{mech}'")

def argmin(args : Sequence[SensitiveInt | SensitiveFloat],
           *,
           eps  : floating | None = None,
           rho  : floating | None = None,
           mech : str             = "exp",
           ) -> int:
    args_negative = [-x for x in args]
    if mech == "exp":
        return exponential_mechanism(args_negative,
                                     eps = float(eps) if eps is not None else None,
                                     rho = float(rho) if rho is not None else None)
    else:
        raise ValueError(f"Unknown DP mechanism: '{mech}'")



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any
import math

from .util import Accountant, ParallelAccountant, SubsamplingAccountant
from .approx import ApproxDPAccountant
from .. import egrpc

PureBudgetType = float

@egrpc.remoteclass
class PureDPAccountant(Accountant[PureBudgetType]):
    @staticmethod
    def family_name() -> str:
        return "pure"

    @egrpc.property
    def budget_spent(self) -> PureBudgetType:
        return self._budget_spent

    def propagate(self, next_budget_spent: PureBudgetType, parent: Accountant[Any]) -> None:
        diff = next_budget_spent - self._budget_spent
        if isinstance(parent, PureDPAccountant):
            parent.spend(diff)
        elif isinstance(parent, PureDPParallelAccountant):
            parent.spend(next_budget_spent)
        elif isinstance(parent, PureDPSubsamplingAccountant):
            parent.spend(next_budget_spent)
        elif isinstance(parent, ApproxDPAccountant):
            parent.spend((diff, 0))
        else:
            raise Exception

    def compose(self, budget1: PureBudgetType, budget2: PureBudgetType) -> PureBudgetType:
        return budget1 + budget2

    def zero(self) -> PureBudgetType:
        return 0.0

    def exceeds(self, budget1: PureBudgetType, budget2: PureBudgetType) -> bool:
        return budget1 > budget2

    def assert_budget(self, budget: PureBudgetType) -> None:
        assert budget >= 0

    @classmethod
    def normalize_budget(cls, budget: Any) -> PureBudgetType | None:
        if budget is None:
            return None
        elif isinstance(budget, (int, float)):
            normalized = float(budget)
            assert normalized >= 0
            return normalized
        else:
            raise TypeError("Pure accountant budget must be a single float value.")

    @staticmethod
    def parallel_accountant() -> type[PureDPParallelAccountant]:
        return PureDPParallelAccountant

    @staticmethod
    def subsampling_accountant() -> type[PureDPSubsamplingAccountant]:
        return PureDPSubsamplingAccountant

class PureDPParallelAccountant(ParallelAccountant[PureBudgetType]):
    @staticmethod
    def family_name() -> str:
        return "pure"

    def propagate(self, next_budget_spent: PureBudgetType, parent: Accountant[Any]) -> None:
        if isinstance(parent, PureDPAccountant):
            parent.spend(next_budget_spent - self._budget_spent)
        else:
            raise Exception

    def compose(self, budget1: PureBudgetType, budget2: PureBudgetType) -> PureBudgetType:
        return max(budget1, budget2)

    def zero(self) -> PureBudgetType:
        return 0.0

    def exceeds(self, budget1: PureBudgetType, budget2: PureBudgetType) -> bool:
        return budget1 > budget2

    def assert_budget(self, budget: PureBudgetType) -> None:
        assert budget >= 0

    @classmethod
    def normalize_budget(cls, budget: Any) -> PureBudgetType | None:
        return PureDPAccountant.normalize_budget(budget)

class PureDPSubsamplingAccountant(SubsamplingAccountant[PureBudgetType]):
    @staticmethod
    def family_name() -> str:
        return "pure"

    def propagate(self, next_budget_spent: PureBudgetType, parent: Accountant[Any]) -> None:
        if isinstance(parent, PureDPAccountant):
            parent.spend(next_budget_spent - self._budget_spent)
        else:
            raise Exception

    def compose(self, budget1: PureBudgetType, budget2: PureBudgetType) -> PureBudgetType:
        q = self._sampling_rate
        amp = math.log(1 + q * (math.exp(budget2) - 1))
        assert budget1 <= amp
        return amp

    def zero(self) -> PureBudgetType:
        return 0.0

    def exceeds(self, budget1: PureBudgetType, budget2: PureBudgetType) -> bool:
        return budget1 > budget2

    def assert_budget(self, budget: PureBudgetType) -> None:
        assert budget >= 0

    @classmethod
    def normalize_budget(cls, budget: Any) -> PureBudgetType | None:
        return PureDPAccountant.normalize_budget(budget)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TypeVar, Any, Sequence
import json

import pandas as _pd

from .. import egrpc
from ..util import DPError, ElementType, realnum
from ..alignment import assert_privacy_axis
from ..realexpr import RealExpr
from ..accountants import BudgetType, Accountant, PureDPAccountant, ApproxDPAccountant
from .domain import CategoryDomain, normalize_column_schema, apply_column_schema, column_schema2domain
from .series import PrivSeries
from .dataframe import PrivDataFrame, SensitiveDataFrame

T = TypeVar("T")

@egrpc.function
def read_csv(filepath     : str,
             schemapath   : str | None        = None,
             *,
             accountant   : str | None        = "approx",
             budget_limit : BudgetType | None = None,
             ) -> PrivDataFrame:
    # TODO: more vaildation for the input data
    df = _pd.read_csv(filepath)

    if schemapath is not None:
        with open(schemapath, "r") as f:
            schema = json.load(f)
    else:
        schema = dict()

    domains = dict()
    user_key = None
    for col in df.columns:
        if not isinstance(col, str):
            raise ValueError("Column name must be a string.")

        if col in schema:
            col_schema = schema[col]
        else:
            col_schema = dict(type="string" if df.dtypes[col] == "object" else df.dtypes[col])

        is_user_key = col_schema["user_key"] if "user_key" in col_schema else False
        if is_user_key:
            user_key = col
            # TODO: handle duplicates of user keys

        col_schema = normalize_column_schema(col_schema)

        df[col] = apply_column_schema(df[col], col_schema, col)

        domains[col] = column_schema2domain(col_schema)

    acc: Accountant[Any]

    if accountant == "pure":
        acc = PureDPAccountant(budget_limit=PureDPAccountant.normalize_budget(budget_limit))
    elif accountant == "approx":
        acc = ApproxDPAccountant(budget_limit=ApproxDPAccountant.normalize_budget(budget_limit))
    else:
        raise ValueError(f"Unknown accountant: '{accountant}'")

    acc.set_as_root(name=filepath)

    # consolidate dataframe blocks (for performance)
    df = df.copy()

    return PrivDataFrame(data       = df,
                         domains    = domains,
                         distance   = RealExpr(1),
                         user_key   = user_key,
                         accountant = acc)

@egrpc.function
def crosstab(index        : PrivSeries[ElementType], # TODO: support Sequence[PrivSeries[ElementType]]
             columns      : PrivSeries[ElementType], # TODO: support Sequence[PrivSeries[ElementType]]
             values       : PrivSeries[ElementType] | None = None,
             rownames     : list[str] | None               = None,
             colnames     : list[str] | None               = None,
             *,
             aggfunc      : None                           = None,
             margins      : bool                           = False,
             margins_name : str                            = "All",
             dropna       : bool                           = True,
             normalize    : bool | str | int               = False, # TODO: support Literal["all", "index", "columns", 0, 1] in egrpc
             ) -> SensitiveDataFrame:
    index._assert_not_uldp()
    columns._assert_not_uldp()

    if normalize is not False:
        # TODO: what is the sensitivity?
        print(normalize)
        raise NotImplementedError

    if values is not None or aggfunc is not None:
        # TODO: hard to accept arbitrary user functions
        raise NotImplementedError

    if margins:
        # Sensitivity must be managed separately for value counts and margins
        raise DPError("`margins=True` is not supported. Please manually calculate margins after adding noise.")

    if isinstance(index.domain, CategoryDomain):
        rowvalues = index.domain.categories
    else:
        raise DPError("Series for crosstab() must be of a categorical type")

    if isinstance(columns.domain, CategoryDomain):
        colvalues = columns.domain.categories
    else:
        raise DPError("Series for crosstab() must be of a categorical type")

    # if not dropna and (not any(_np.isnan(rowvalues)) or not any(_np.isnan(colvalues))):
    #     # TODO: consider handling for pd.NA
    #     warnings.warn("Counts for NaN will be dropped from the result because NaN is not included in `rowvalues`/`colvalues`", UserWarning)

    assert_privacy_axis(index, columns)

    counts = _pd.crosstab(index._value, columns._value,
                          values=None, rownames=rownames, colnames=colnames,
                          aggfunc=None, margins=False, margins_name=margins_name,
                          dropna=dropna, normalize=False)

    # Select only the specified values and fill non-existent counts with 0
    counts = counts.reindex(list(rowvalues), axis="index") \
                   .reindex(list(colvalues), axis="columns") \
                   .fillna(0).astype(int)

    return SensitiveDataFrame(data                = counts,
                              distance_group_axes = None,
                              distance            = index._distance,
                              parents             = [index, columns])

# TODO: change multifunction -> function by type checking in egrpc.function
@egrpc.multifunction
def cut(x              : PrivSeries[Any],
        bins           : Sequence[realnum],
        right          : bool                                = True,
        labels         : Sequence[ElementType] | bool | None = None,
        retbins        : bool                                = False,
        precision      : int                                 = 3,
        include_lowest : bool                                = False
        # TODO: add more parameters
        ) -> PrivSeries[Any]:
    x._assert_not_uldp()

    ser = _pd.cut(x._value, bins=bins, right=right, labels=labels, retbins=retbins, precision=precision, include_lowest=include_lowest) # type: ignore

    new_domain = CategoryDomain(categories=list(ser.dtype.categories))

    return PrivSeries[Any](data           = ser,
                           domain         = new_domain,
                           distance       = x._distance,
                           parents        = [x],
                           keep_alignment = True)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any, Sequence, TypeGuard

import numpy as _np
import numpy.typing as _npt

from .. import egrpc
from ..util import DPError, realnum
from ..realexpr import RealExpr
from .array import PrivNDArray, SensitiveNDArray
from .domain import NDArrayDomain, ValueRange

def _is_sequence(value: Any) -> TypeGuard[Sequence[Any]]:
    if isinstance(value, _np.ndarray):
        return value.ndim > 0
    return isinstance(value, Sequence)

@egrpc.function
def histogram(a     : PrivNDArray,
              bins  : int | Sequence[realnum]        = 10,
              range : tuple[realnum, realnum] | None = None,
              ) -> tuple[SensitiveNDArray, _npt.NDArray[_np.floating[Any]]]:
    if a.ndim != 1:
        if not all(dim == 1 for dim in a._value.shape[1:]):
            raise ValueError("`a` must be 1-D or have trailing singleton dimensions.")

    if _is_sequence(bins):
        if range is not None:
            raise ValueError("`range` must be None when explicit bin edges are provided.")
    else:
        if range is None:
            raise DPError("`range` must be specified when bins are given as counts.")

    hist, edges = _np.histogram(a._value, bins=bins, range=range) # type: ignore

    sensitive_hist = SensitiveNDArray(value     = hist,
                                      distance  = a._distance,
                                      norm_type = "l1",
                                      parents   = [a])

    return sensitive_hist, edges

@egrpc.function
def histogramdd(sample : PrivNDArray,
                bins   : int | Sequence[int] | Sequence[Sequence[realnum]] = 10,
                range  : Sequence[tuple[realnum, realnum]] | None          = None,
                ) -> tuple[SensitiveNDArray, tuple[_npt.NDArray[_np.floating[Any]], ...]]:
    if sample.ndim == 1:
        dim = 1
    elif sample.ndim == 2:
        dim = sample._value.shape[1]
    else:
        raise ValueError("`sample` must be (N, D) array")

    if _is_sequence(bins) and len(bins) > 0 and _is_sequence(bins[0]):
        if range is not None:
            raise ValueError("`range` must be None when explicit bin edges are provided.")
        if len(bins) != dim:
            raise ValueError("`bins` length must match the dimensionality of the sample.")
    else:
        if range is None:
            raise DPError("`range` must be specified when bins are given as counts.")
        if len(range) != dim:
            raise ValueError("`range` length must match the dimensionality of the sample.")
        if _is_sequence(bins) and len(bins) != dim:
            raise ValueError("`bins` length must match the dimensionality of the sample.")

    hist, edges = _np.histogramdd(sample._value, bins=bins, range=range) # type: ignore

    sensitive_hist = SensitiveNDArray(value     = hist,
                                      distance  = sample._distance,
                                      norm_type = "l1",
                                      parents   = [sample])

    return sensitive_hist, edges

def _maximum_value_ranges(vr1: ValueRange, vr2: ValueRange) -> ValueRange:
    if vr1 is None or vr2 is None:
        return None
    lo1, hi1 = vr1
    lo2, hi2 = vr2
    new_lo = max(lo1, lo2) if lo1 is not None and lo2 is not None else None
    new_hi = max(hi1, hi2) if hi1 is not None and hi2 is not None else None
    return (new_lo, new_hi)

def _maximum_value_range_scalar(vr: ValueRange, c: float) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    new_lo = max(lo, c) if lo is not None else None
    new_hi = max(hi, c) if hi is not None else None
    return (new_lo, new_hi)

def _minimum_value_ranges(vr1: ValueRange, vr2: ValueRange) -> ValueRange:
    if vr1 is None or vr2 is None:
        return None
    lo1, hi1 = vr1
    lo2, hi2 = vr2
    new_lo = min(lo1, lo2) if lo1 is not None and lo2 is not None else None
    new_hi = min(hi1, hi2) if hi1 is not None and hi2 is not None else None
    return (new_lo, new_hi)

def _minimum_value_range_scalar(vr: ValueRange, c: float) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    new_lo = min(lo, c) if lo is not None else None
    new_hi = min(hi, c) if hi is not None else None
    return (new_lo, new_hi)

def _exp_value_range(vr: ValueRange) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    new_lo = _np.exp(lo) if lo is not None else None
    new_hi = _np.exp(hi) if hi is not None else None
    return (new_lo, new_hi)

def _log_value_range(vr: ValueRange) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    if lo is not None and lo <= 0:
        return None
    new_lo = _np.log(lo) if lo is not None else None
    new_hi = _np.log(hi) if hi is not None else None
    return (new_lo, new_hi)

@egrpc.function
def maximum(x1: PrivNDArray, x2: PrivNDArray | realnum) -> PrivNDArray:
    if isinstance(x2, PrivNDArray):
        x1._check_axis_aligned(x2)
        new_vr = _maximum_value_ranges(x1.domain.value_range, x2.domain.value_range)
        return PrivNDArray(value          = _np.maximum(x1._value, x2._value),
                           distance       = x1._distance,
                           privacy_axis   = x1._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [x1, x2],
                           keep_alignment = True)
    else:
        new_vr = _maximum_value_range_scalar(x1.domain.value_range, float(x2))
        return PrivNDArray(value          = _np.maximum(x1._value, x2),
                           distance       = x1._distance,
                           privacy_axis   = x1._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [x1],
                           keep_alignment = True)

@egrpc.function
def minimum(x1: PrivNDArray, x2: PrivNDArray | realnum) -> PrivNDArray:
    if isinstance(x2, PrivNDArray):
        x1._check_axis_aligned(x2)
        new_vr = _minimum_value_ranges(x1.domain.value_range, x2.domain.value_range)
        return PrivNDArray(value          = _np.minimum(x1._value, x2._value),
                           distance       = x1._distance,
                           privacy_axis   = x1._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [x1, x2],
                           keep_alignment = True)
    else:
        new_vr = _minimum_value_range_scalar(x1.domain.value_range, float(x2))
        return PrivNDArray(value          = _np.minimum(x1._value, x2),
                           distance       = x1._distance,
                           privacy_axis   = x1._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [x1],
                           keep_alignment = True)

@egrpc.function
def exp(x: PrivNDArray) -> PrivNDArray:
    new_vr = _exp_value_range(x.domain.value_range)
    return PrivNDArray(value          = _np.exp(x._value),
                       distance       = x._distance,
                       privacy_axis   = x._privacy_axis,
                       domain         = NDArrayDomain(value_range=new_vr),
                       parents        = [x],
                       keep_alignment = True)

@egrpc.function
def log(x: PrivNDArray) -> PrivNDArray:
    new_vr = _log_value_range(x.domain.value_range)
    return PrivNDArray(value          = _np.log(x._value),
                       distance       = x._distance,
                       privacy_axis   = x._privacy_axis,
                       domain         = NDArrayDomain(value_range=new_vr),
                       parents        = [x],
                       keep_alignment = True)

@egrpc.function
def eye(N: int, M: int | None = None, k: int = 0) -> SensitiveNDArray:
    # dummy SensitiveNDArray for constructing one hot vectors (np.eye(N)[y])
    return SensitiveNDArray(value=_np.eye(N, M, k), distance=RealExpr(0))

def _union_value_ranges(ranges: Sequence[ValueRange]) -> ValueRange:
    if not ranges or any(r is None or r[0] is None or r[1] is None for r in ranges):
        return None
    los = [r[0] for r in ranges if r is not None and r[0] is not None]
    his = [r[1] for r in ranges if r is not None and r[1] is not None]
    return (min(los), max(his))

@egrpc.function
def concatenate(arrays: Sequence[PrivNDArray], axis: int = 0) -> PrivNDArray:
    if len(arrays) == 0:
        raise ValueError("need at least one array to concatenate")

    first = arrays[0]
    privacy_axis = first._privacy_axis
    alignment_signature = first.alignment_signature

    if axis < 0:
        axis = first.ndim + axis

    for arr in arrays[1:]:
        if arr.alignment_signature != alignment_signature:
            raise DPError("All arrays must have the same alignment_signature.")
        if arr._privacy_axis != privacy_axis:
            raise DPError("All arrays must have the same privacy_axis.")

    if axis == privacy_axis:
        # TODO: support concatenation along privacy_axis by summing distances
        raise DPError("Cannot concatenate along privacy_axis.")

    new_vr = _union_value_ranges([arr.domain.value_range for arr in arrays])

    result = _np.concatenate([arr._value for arr in arrays], axis=axis)
    return PrivNDArray(value          = result,
                       distance       = first._distance,
                       privacy_axis   = privacy_axis,
                       domain         = NDArrayDomain(value_range=new_vr),
                       parents        = list(arrays),
                       keep_alignment = True)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TypeVar, Generic, Any, overload, Iterable, cast, Sequence, Iterator
import contextlib

import numpy as _np

from .util import integer, floating, realnum, is_integer, is_floating, DPError
from .realexpr import RealExpr, _max as dmax
from .accountants import Accountant, ParallelAccountant, DummyAccountant, PureDPAccountant, ApproxDPAccountant, zCDPAccountant, RDPAccountant, get_lsca_of_same_family, BudgetType, AccountingGroup
from . import egrpc

T = TypeVar("T")

@egrpc.remoteclass
class Prisoner(Generic[T]):
    _value            : T
    _distance         : RealExpr
    _accountant       : Accountant[Any]
    _accounting_group : AccountingGroup | None

    def __init__(self,
                 value            : T,
                 distance         : RealExpr,
                 *,
                 parents          : Sequence[Prisoner[Any]] = [], # FIXME: do we really need parents here?
                 accountant       : Accountant[Any] | None  = None,
                 accounting_group : AccountingGroup | None  = None,
                 ):
        self._value    = value
        self._distance = distance

        if distance.is_zero():
            # constant (public value)
            self._accountant       = DummyAccountant()
            self._accounting_group = None

        elif len(parents) == 0:
            # root prisoner
            if accountant is None:
                raise ValueError("accountant must be set for the root prisoner.")
            self._accountant       = accountant
            self._accounting_group = accounting_group if accounting_group is not None else accountant.accounting_group

        elif accountant is not None:
            # accountant is explicitly specified
            if len(parents) != 1:
                raise ValueError("different accountant cannot be set for a prisoner with multiple parents")
            self._accountant       = accountant
            self._accounting_group = accounting_group if accounting_group is not None else accountant.accounting_group

        elif len(parents) == 1:
            # single parent
            self._accountant       = parents[0].accountant
            self._accounting_group = parents[0]._accounting_group

        else:
            # multiple parents
            lsca_accountant        = get_lsca_of_same_family([p.accountant for p in parents if not isinstance(p.accountant, DummyAccountant)])
            self._accountant       = lsca_accountant.get_parent() if isinstance(lsca_accountant, ParallelAccountant) else lsca_accountant
            self._accounting_group = self._accountant.accounting_group

    @egrpc.property
    def accountant(self) -> Accountant[Any]:
        return self._accountant

    @egrpc.method
    def set_accountant(self, accountant: Accountant[Any]) -> None:
        if self._accounting_group is None:
            self._accountant = accountant
            return
        acc_group = accountant.accounting_group
        if acc_group is None or not acc_group.is_ancestor_or_same(self._accounting_group):
            raise DPError("accountant's accounting_group must be an ancestor or the same as prisoner's accounting_group")
        self._accountant = accountant

    @property
    def accounting_group(self) -> AccountingGroup | None:
        return self._accounting_group

    def __str__(self) -> str:
        return "<***>"

    def __repr__(self) -> str:
        return "<***>"

    @egrpc.property
    def max_distance(self) -> realnum:
        return self._distance.max()

@egrpc.remoteclass
class SensitiveInt(Prisoner[integer]):
    def __init__(self,
                 value      : integer,
                 distance   : RealExpr                = RealExpr(0),
                 *,
                 parents    : Sequence[Prisoner[Any]] = [],
                 accountant : Accountant[Any] | None  = None,
                 ):
        if not is_integer(value):
            raise ValueError("`value` must be int for SensitveInt.")
        super().__init__(value, distance, parents=parents, accountant=accountant)

    def __str__(self) -> str:
        return "<*** (int)>"

    def __repr__(self) -> str:
        return "<*** (int)>"

    @egrpc.method
    def __neg__(self) -> SensitiveInt:
        return SensitiveInt(-self._value, distance=self._distance, parents=[self])

    @egrpc.multimethod
    def __add__(self, other: integer) -> SensitiveInt:
        return SensitiveInt(self._value + other, distance=self._distance, parents=[self])

    @__add__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(self._value + other, distance=self._distance, parents=[self])

    @__add__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(self._value + other._value, distance=self._distance + other._distance, parents=[self, other])

    @__add__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value + other._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __radd__(self, other: integer) -> SensitiveInt: # type: ignore[misc]
        return SensitiveInt(other + self._value, distance=self._distance, parents=[self])

    @__radd__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(other + self._value, distance=self._distance, parents=[self])

    @__radd__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(other._value + self._value, distance=self._distance + other._distance, parents=[self, other])

    @__radd__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value + self._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __sub__(self, other: integer) -> SensitiveInt:
        return SensitiveInt(self._value - other, distance=self._distance, parents=[self])

    @__sub__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(self._value - other, distance=self._distance, parents=[self])

    @__sub__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(self._value - other._value, distance=self._distance + other._distance, parents=[self, other])

    @__sub__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value - other._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __rsub__(self, other: integer) -> SensitiveInt: # type: ignore[misc]
        return SensitiveInt(other - self._value, distance=self._distance, parents=[self])

    @__rsub__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(other - self._value, distance=self._distance, parents=[self])

    @__rsub__.register
    def _(self, other: SensitiveInt) -> SensitiveInt:
        return SensitiveInt(other._value - self._value, distance=self._distance + other._distance, parents=[self, other])

    @__rsub__.register
    def _(self, other: SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value - self._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __mul__(self, other: integer) -> SensitiveInt:
        return SensitiveInt(self._value * other, distance=self._distance * _np.abs(other), parents=[self])

    @__mul__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(self._value * other, distance=self._distance * _np.abs(other), parents=[self])

    @egrpc.multimethod
    def __rmul__(self, other: integer) -> SensitiveInt: # type: ignore[misc]
        return SensitiveInt(other * self._value, distance=self._distance * _np.abs(other), parents=[self])

    @__rmul__.register
    def _(self, other: floating) -> SensitiveFloat:
        return SensitiveFloat(other * self._value, distance=self._distance * _np.abs(other), parents=[self])

    def reveal(self,
               *,
               eps   : floating | None = None,
               delta : floating | None = None,
               rho   : floating | None = None,
               scale : floating | None = None,
               mech  : str             = "laplace",
               ) -> float:
        if mech == "laplace":
            from .mechanism import laplace_mechanism
            return laplace_mechanism(self,
                                     eps   = float(eps)   if eps   is not None else None,
                                     scale = float(scale) if scale is not None else None)
        elif mech == "gaussian":
            from .mechanism import gaussian_mechanism
            return gaussian_mechanism(self,
                                      eps   = float(eps)   if eps   is not None else None,
                                      delta = float(delta) if delta is not None else None,
                                      rho   = float(rho)   if rho   is not None else None,
                                      scale = float(scale) if scale is not None else None)
        else:
            raise ValueError(f"Unknown DP mechanism: '{mech}'")

@egrpc.remoteclass
class SensitiveFloat(Prisoner[floating]):
    def __init__(self,
                 value      : floating,
                 distance   : RealExpr                = RealExpr(0),
                 *,
                 parents    : Sequence[Prisoner[Any]] = [],
                 accountant : Accountant[Any] | None  = None,
                 ):
        if not is_floating(value):
            raise ValueError("`value` must be float for SensitveFloat.")
        super().__init__(value, distance, parents=parents, accountant=accountant)

    def __str__(self) -> str:
        return "<*** (float)>"

    def __repr__(self) -> str:
        return "<*** (float)>"

    @egrpc.method
    def __neg__(self) -> SensitiveFloat:
        return SensitiveFloat(-self._value, distance=self._distance, parents=[self])

    @egrpc.multimethod
    def __add__(self, other: realnum) -> SensitiveFloat:
        return SensitiveFloat(self._value + other, distance=self._distance, parents=[self])

    @__add__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value + other._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __radd__(self, other: realnum) -> SensitiveFloat: # type: ignore[misc]
        return SensitiveFloat(other + self._value, distance=self._distance, parents=[self])

    @__radd__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value + self._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __sub__(self, other: realnum) -> SensitiveFloat:
        return SensitiveFloat(self._value - other, distance=self._distance, parents=[self])

    @__sub__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(self._value - other._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __rsub__(self, other: realnum) -> SensitiveFloat: # type: ignore[misc]
        return SensitiveFloat(other - self._value, distance=self._distance, parents=[self])

    @__rsub__.register
    def _(self, other: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
        return SensitiveFloat(other._value - self._value, distance=self._distance + other._distance, parents=[self, other])

    @egrpc.multimethod
    def __mul__(self, other: realnum) -> SensitiveFloat:
        return SensitiveFloat(self._value * other, distance=self._distance * _np.abs(other), parents=[self])

    @egrpc.multimethod
    def __rmul__(self, other: realnum) -> SensitiveFloat: # type: ignore[misc]
        return SensitiveFloat(other * self._value, distance=self._distance * _np.abs(other), parents=[self])

    def reveal(self,
               *,
               eps   : floating | None = None,
               delta : floating | None = None,
               rho   : floating | None = None,
               scale : floating | None = None,
               mech  : str             = "laplace",
               ) -> float:
        if mech == "laplace":
            from .mechanism import laplace_mechanism
            return laplace_mechanism(self,
                                     eps   = float(eps)   if eps   is not None else None,
                                     scale = float(scale) if scale is not None else None)
        elif mech == "gaussian":
            from .mechanism import gaussian_mechanism
            return gaussian_mechanism(self,
                                      eps   = float(eps)   if eps   is not None else None,
                                      delta = float(delta) if delta is not None else None,
                                      rho   = float(rho)   if rho   is not None else None,
                                      scale = float(scale) if scale is not None else None)
        else:
            raise ValueError(f"Unknown DP mechanism: '{mech}'")

@egrpc.multifunction
def max2(a: SensitiveInt | SensitiveFloat, b: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
    return SensitiveFloat(max(float(a._value), float(b._value)), distance=dmax(a._distance, b._distance), parents=[a, b])

@max2.register
def _(a: SensitiveInt, b: SensitiveInt) -> SensitiveInt:
    return SensitiveInt(max(int(a._value), int(b._value)), distance=dmax(a._distance, b._distance), parents=[a, b])

@overload
def _max(*args: SensitiveInt) -> SensitiveInt: ...
@overload
def _max(*args: SensitiveFloat) -> SensitiveFloat: ...
@overload
def _max(*args: Iterable[SensitiveInt]) -> SensitiveInt: ...
@overload
def _max(*args: Iterable[SensitiveFloat]) -> SensitiveFloat: ...
@overload
def _max(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat: ...

def _max(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat:
    if len(args) == 0:
        raise TypeError("max() expected at least one argment.")

    if len(args) == 1:
        if not isinstance(args[0], Iterable):
            raise TypeError("The first arg passed to max() is not iterable.")
        iterable = args[0]
    else:
        iterable = cast(tuple[SensitiveInt | SensitiveFloat, ...], args)

    it = iter(iterable)
    try:
        result = next(it)
    except StopIteration:
        raise ValueError("List passed to max() is empty.")

    for x in it:
        result = max2(result, x)

    return result

@egrpc.multifunction
def min2(a: SensitiveInt | SensitiveFloat, b: SensitiveInt | SensitiveFloat) -> SensitiveFloat:
    return SensitiveFloat(min(float(a._value), float(b._value)), distance=dmax(a._distance, b._distance), parents=[a, b])

@min2.register
def _(a: SensitiveInt, b: SensitiveInt) -> SensitiveInt:
    return SensitiveInt(min(int(a._value), int(b._value)), distance=dmax(a._distance, b._distance), parents=[a, b])

@overload
def _min(*args: SensitiveInt) -> SensitiveInt: ...
@overload
def _min(*args: SensitiveFloat) -> SensitiveFloat: ...
@overload
def _min(*args: Iterable[SensitiveInt]) -> SensitiveInt: ...
@overload
def _min(*args: Iterable[SensitiveFloat]) -> SensitiveFloat: ...
@overload
def _min(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat: ...

def _min(*args: Iterable[SensitiveInt | SensitiveFloat] | SensitiveInt | SensitiveFloat) -> SensitiveInt | SensitiveFloat:
    if len(args) == 0:
        raise TypeError("min() expected at least one argment.")

    if len(args) == 1:
        if not isinstance(args[0], Iterable):
            raise TypeError("The first arg passed to min() is not iterable.")
        iterable = args[0]
    else:
        iterable = cast(tuple[SensitiveInt | SensitiveFloat, ...], args)

    it = iter(iterable)
    try:
        result = next(it)
    except StopIteration:
        raise ValueError("List passed to min() is empty.")

    for x in it:
        result = min2(result, x)

    return result

@egrpc.function
def create_accountant(accountant_type : str,
                      *,
                      parent          : Accountant[Any],
                      budget_limit    : BudgetType | None        = None,
                      delta           : realnum | None           = None,
                      alpha           : Sequence[realnum] | None = None,
                      prepaid         : bool                     = False,
                      ) -> Accountant[Any]:
    if prepaid and budget_limit is None:
        raise ValueError("prepaid=True requires budget_limit to be specified")

    accountant_type_lower = accountant_type.lower()

    # zCDP/RDP odometer (non-prepaid mode) is not implemented
    if accountant_type_lower in ("zcdp", "rdp"):
        if budget_limit is None or not prepaid:
            raise NotImplementedError(
                f"{accountant_type_lower.upper()} odometer is not implemented. "
                "Use prepaid=True with budget_limit to use filter mode."
            )

    # budget_limit creates an intermediate accountant of the parent's type
    if budget_limit is not None:
        parent = type(parent)(budget_limit=budget_limit, parent=parent, prepaid=prepaid)
        if delta is None and isinstance(budget_limit, tuple):
            delta = budget_limit[1]

    delta_float = float(delta) if delta is not None else None
    if accountant_type_lower == "pure":
        return PureDPAccountant(parent=parent)
    elif accountant_type_lower == "approx":
        return ApproxDPAccountant(parent=parent)
    elif accountant_type_lower == "zcdp":
        return zCDPAccountant(parent=parent, delta=delta_float)
    elif accountant_type_lower == "rdp":
        assert alpha is not None
        alpha_float = [float(a) for a in alpha]
        return RDPAccountant(alpha=alpha_float, parent=parent, delta=delta_float)
    else:
        raise ValueError(f"Unknown DP type: {accountant_type}. Expected 'pure', 'approx', 'zcdp', or 'rdp'.")

P = TypeVar("P", bound=Prisoner[Any])

@contextlib.contextmanager
def _dp_context(prisoners       : P | Sequence[P],
                accountant_type : str,
                budget_limit    : BudgetType | None        = None,
                delta           : realnum | None           = None,
                alpha           : Sequence[realnum] | None = None,
                prepaid         : bool                     = False,
                ) -> Iterator[P | Sequence[P]]:
    if isinstance(prisoners, Prisoner):
        prisoner_list: list[P] = [prisoners]  # type: ignore[list-item]
    else:
        prisoner_list = list(prisoners)

    if len(prisoner_list) == 0:
        raise ValueError("At least one prisoner is required")

    # Check all prisoners have the same accountant
    first_accountant = prisoner_list[0].accountant
    for p in prisoner_list[1:]:
        if p.accountant is not first_accountant:
            raise DPError("All prisoners must have the same accountant")

    old_accountants = [p.accountant for p in prisoner_list]
    new_accountant = create_accountant(accountant_type, parent=first_accountant, budget_limit=budget_limit, delta=delta, alpha=alpha, prepaid=prepaid)

    for p in prisoner_list:
        p.set_accountant(new_accountant)
    try:
        yield prisoners  # type: ignore[misc]
    finally:
        for p, old_acc in zip(prisoner_list, old_accountants):
            p.set_accountant(old_acc)

@contextlib.contextmanager
def pureDP(prisoners    : P | Sequence[P],
           budget_limit : BudgetType | None = None,
           prepaid      : bool              = False,
           ) -> Iterator[P | Sequence[P]]:
    with _dp_context(prisoners, "pure", budget_limit=budget_limit, prepaid=prepaid) as ctx:
        yield ctx

@contextlib.contextmanager
def approxDP(prisoners    : P | Sequence[P],
             budget_limit : BudgetType | None = None,
             prepaid      : bool              = False,
             ) -> Iterator[P | Sequence[P]]:
    with _dp_context(prisoners, "approx", budget_limit=budget_limit, prepaid=prepaid) as ctx:
        yield ctx

@contextlib.contextmanager
def zCDP(prisoners    : P | Sequence[P],
         budget_limit : BudgetType | None = None,
         delta        : realnum | None    = None,
         prepaid      : bool              = False,
         ) -> Iterator[P | Sequence[P]]:
    with _dp_context(prisoners, "zcdp", budget_limit=budget_limit, delta=delta, prepaid=prepaid) as ctx:
        yield ctx

@contextlib.contextmanager
def RDP(prisoners    : P | Sequence[P],
        alpha        : Sequence[realnum],
        budget_limit : BudgetType | None = None,
        delta        : realnum | None    = None,
        prepaid      : bool              = False,
        ) -> Iterator[P | Sequence[P]]:
    with _dp_context(prisoners, "rdp", budget_limit=budget_limit, delta=delta, alpha=alpha, prepaid=prepaid) as ctx:
        yield ctx



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any, Sequence

import numpy as _np
import numpy.typing as _npt

from .. import egrpc
from ..util import DPError, floating, realnum
from ..array_base import PrivArrayBase, SensitiveDimInt
from ..realexpr import RealExpr
from ..accountants import Accountant
from ..prisoner import Prisoner, SensitiveFloat
from .domain import NDArrayDomain, ValueRange
from .util import PrivShape, NDIndex, infer_missing_dim, check_broadcast_privacy_axis

def _negate_value_range(vr: ValueRange) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    new_lo = -hi if hi is not None else None
    new_hi = -lo if lo is not None else None
    return (new_lo, new_hi)

def _shift_value_range(vr: ValueRange, c: float) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    new_lo = lo + c if lo is not None else None
    new_hi = hi + c if hi is not None else None
    return (new_lo, new_hi)

def _rshift_value_range(c: float, vr: ValueRange) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    new_lo = c - hi if hi is not None else None
    new_hi = c - lo if lo is not None else None
    return (new_lo, new_hi)

def _rdiv_value_range(c: float, vr: ValueRange) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    if lo is None or hi is None:
        return None
    if lo <= 0 <= hi:
        return None
    if c == 0:
        return (0.0, 0.0)
    val1 = c / lo
    val2 = c / hi
    return (min(val1, val2), max(val1, val2))

def _scale_value_range(vr: ValueRange, c: float) -> ValueRange:
    if vr is None:
        return None
    lo, hi = vr
    if c >= 0:
        new_lo = lo * c if lo is not None else None
        new_hi = hi * c if hi is not None else None
    else:
        new_lo = hi * c if hi is not None else None
        new_hi = lo * c if lo is not None else None
    return (new_lo, new_hi)

def _add_value_ranges(vr1: ValueRange, vr2: ValueRange) -> ValueRange:
    if vr1 is None or vr2 is None:
        return None
    lo1, hi1 = vr1
    lo2, hi2 = vr2
    new_lo = lo1 + lo2 if lo1 is not None and lo2 is not None else None
    new_hi = hi1 + hi2 if hi1 is not None and hi2 is not None else None
    return (new_lo, new_hi)

def _sub_value_ranges(vr1: ValueRange, vr2: ValueRange) -> ValueRange:
    if vr1 is None or vr2 is None:
        return None
    lo1, hi1 = vr1
    lo2, hi2 = vr2
    new_lo = lo1 - hi2 if lo1 is not None and hi2 is not None else None
    new_hi = hi1 - lo2 if hi1 is not None and lo2 is not None else None
    return (new_lo, new_hi)

def _mul_value_ranges(vr1: ValueRange, vr2: ValueRange) -> ValueRange:
    if vr1 is None or vr2 is None:
        return None
    lo1, hi1 = vr1
    lo2, hi2 = vr2
    if lo1 is None or hi1 is None or lo2 is None or hi2 is None:
        return None
    products = [lo1 * lo2, lo1 * hi2, hi1 * lo2, hi1 * hi2]
    return (min(products), max(products))

def _div_value_ranges(vr1: ValueRange, vr2: ValueRange) -> ValueRange:
    if vr1 is None or vr2 is None:
        return None
    lo1, hi1 = vr1
    lo2, hi2 = vr2
    if lo1 is None or hi1 is None or lo2 is None or hi2 is None:
        return None
    if lo2 <= 0 <= hi2:
        return None
    quotients = [lo1 / lo2, lo1 / hi2, hi1 / lo2, hi1 / hi2]
    return (min(quotients), max(quotients))


@egrpc.remoteclass
class PrivNDArray(PrivArrayBase[_npt.NDArray[_np.floating[Any]]]):
    __array_priority__ = 1000  # To prevent NumPy from treating PrivNDArray as a scalar operand
    _domain: NDArrayDomain

    def __init__(self,
                 value          : Any,
                 distance       : RealExpr,
                 privacy_axis   : int,
                 domain         : NDArrayDomain | None         = None,
                 *,
                 parents        : Sequence[PrivArrayBase[Any]] = [],
                 accountant     : Accountant[Any] | None       = None,
                 keep_alignment : bool                         = False,
                 ) -> None:
        array = _np.asarray(value)
        if not _np.issubdtype(array.dtype, _np.number):
            raise TypeError("PrivNDArray requires a numeric dtype.")

        array.setflags(write=False) # make it immutable to avoid unexpected write to the same view

        self._domain = domain if domain is not None else NDArrayDomain()

        if privacy_axis < 0:
            privacy_axis += array.ndim
        if not 0 <= privacy_axis < array.ndim:
            raise ValueError("privacy_axis is out of bounds for the array dimension.")

        super().__init__(value          = array,
                         distance       = distance,
                         privacy_axis   = privacy_axis,
                         parents        = list(parents),
                         accountant     = accountant,
                         keep_alignment = keep_alignment)

    @egrpc.property
    def shape(self) -> PrivShape:
        return tuple(
            SensitiveDimInt(value               = int(dim),
                            distance            = self._distance,
                            alignment_signature = self._alignment_signature,
                            scale               = 1,
                            parents             = [self])
            if i == self._privacy_axis else int(dim)
            for i, dim in enumerate(self._value.shape)
        )

    @egrpc.property
    def ndim(self) -> int:
        return self._value.ndim

    @egrpc.property
    def domain(self) -> NDArrayDomain:
        return self._domain

    @egrpc.property
    def T(self) -> PrivNDArray:
        return self.transpose()

    @egrpc.method
    def transpose(self, axes: tuple[int, ...] | list[int] | None = None) -> PrivNDArray:
        if axes is None:
            axes_tuple = tuple(reversed(range(self.ndim)))
        else:
            axes_tuple = tuple(axes)
            if len(axes_tuple) != self.ndim:
                raise ValueError("axes tuple must include each axis exactly once.")

        if sorted(axes_tuple) != list(range(self.ndim)):
            raise ValueError("axes must be a permutation of current axes.")

        return PrivNDArray(value          = self._value.transpose(axes_tuple),
                           distance       = self._distance,
                           privacy_axis   = axes_tuple.index(self._privacy_axis),
                           domain         = self._domain,
                           parents        = [self],
                           keep_alignment = True)

    @egrpc.method
    def swapaxes(self, axis1: int, axis2: int) -> PrivNDArray:
        axis1 = axis1 + self.ndim if axis1 < 0 else axis1
        axis2 = axis2 + self.ndim if axis2 < 0 else axis2
        if not (0 <= axis1 < self.ndim) or not (0 <= axis2 < self.ndim):
            raise ValueError("axis index out of bounds.")

        if axis1 == self._privacy_axis:
            new_privacy_axis = axis2
        elif axis2 == self._privacy_axis:
            new_privacy_axis = axis1
        else:
            new_privacy_axis = self._privacy_axis

        return PrivNDArray(value          = self._value.swapaxes(axis1, axis2),
                           distance       = self._distance,
                           privacy_axis   = new_privacy_axis,
                           domain         = self._domain,
                           parents        = [self],
                           keep_alignment = True)

    def reshape(self,
                *shape : int | SensitiveDimInt | PrivShape,
                order  : str = "C",
                ) -> PrivNDArray:
        if order != "C":
            raise ValueError("Only order='C' is supported for reshape.")

        if len(shape) == 1 and isinstance(shape[0], tuple):
            shape_tuple: PrivShape = shape[0]
        else:
            shape_tuple = tuple(d for d in shape if not isinstance(d, tuple))

        return self._reshape_impl(shape_tuple)

    @egrpc.method
    def _reshape_impl(self, shape: PrivShape) -> PrivNDArray:
        output_shape = infer_missing_dim(self.shape, shape)

        priv_dims = [(i, d) for i, d in enumerate(output_shape) if isinstance(d, SensitiveDimInt)]
        if len(priv_dims) != 1:
            raise ValueError("Output shape must contain exactly one SensitiveDimInt.")
        output_axis, output_priv_dim = priv_dims[0]

        scale = output_priv_dim.scale
        for d in output_shape:
            val = d.scale if isinstance(d, SensitiveDimInt) else d
            if val <= 0:
                raise ValueError("All dimensions and scale must be positive for reshape.")

        if output_priv_dim.alignment_signature != self.alignment_signature:
            raise DPError("SensitiveDimInt's alignment_signature does not match the array's alignment_signature.")

        input_axis = self._privacy_axis

        P_in = 1
        for d in self.shape[:input_axis]:
            assert isinstance(d, int)
            P_in *= d

        Q_in = 1
        for d in self.shape[input_axis + 1:]:
            assert isinstance(d, int)
            Q_in *= d

        P_out = 1
        for d in output_shape[:output_axis]:
            assert isinstance(d, int)
            P_out *= d

        Q_out = 1
        for d in output_shape[output_axis + 1:]:
            assert isinstance(d, int)
            Q_out *= d

        if not ((P_in == P_out and Q_in == Q_out * scale) or
                (P_in == P_out * scale and Q_in == Q_out)):
            raise DPError("Reshape would mix data across individuals.")

        final_shape = tuple(
            int(d._value) if isinstance(d, SensitiveDimInt) else d
            for d in output_shape
        )
        reshaped_arr = self._value.reshape(final_shape, order="C")

        return PrivNDArray(value          = reshaped_arr,
                           distance       = self._distance * scale,
                           privacy_axis   = output_axis,
                           domain         = self._domain,
                           parents        = [self],
                           keep_alignment = (scale == 1))

    @egrpc.method
    def sum(self,
            axis     : int | None = None,
            keepdims : bool       = False,
            ) -> SensitiveFloat | SensitiveNDArray | PrivNDArray:
        if axis is None or axis == self._privacy_axis:
            # Sum along privacy_axis (or all axes) -> returns SensitiveFloat or SensitiveNDArray
            if self._domain.norm_bound is None:
                raise DPError("Norm bound is not set. Use clip_norm() before summing along the distance axis.")

            new_distance = self._distance * float(self._domain.norm_bound)
            result = self._value.sum(axis=axis, keepdims=keepdims)

            if result.ndim == 0:
                return SensitiveFloat(float(result), distance=new_distance, parents=[self])
            else:
                return SensitiveNDArray(value     = result,
                                        distance  = new_distance,
                                        norm_type = self._domain.norm_type,
                                        parents   = [self])

        else:
            # Sum along non-distance axis -> returns PrivNDArray
            if axis < 0:
                axis = self.ndim + axis

            if not (0 <= axis < self.ndim):
                raise ValueError(f"axis {axis} is out of bounds for array of dimension {self.ndim}.")

            if keepdims:
                new_privacy_axis = self._privacy_axis
            else:
                if axis < self._privacy_axis:
                    new_privacy_axis = self._privacy_axis - 1
                else:
                    new_privacy_axis = self._privacy_axis

            result = self._value.sum(axis=axis, keepdims=keepdims)

            # Update value_range: sum of N elements scales the range by N
            new_vr: ValueRange = None
            if self.domain.value_range is not None:
                lo, hi = self.domain.value_range
                n = self.shape[axis]
                assert isinstance(n, int)
                new_lo = lo * n if lo is not None else None
                new_hi = hi * n if hi is not None else None
                new_vr = (new_lo, new_hi)

            return PrivNDArray(value          = result,
                               distance       = self._distance,
                               privacy_axis   = new_privacy_axis,
                               domain         = NDArrayDomain(value_range=new_vr),
                               parents        = [self],
                               keep_alignment = True)

    @egrpc.method
    def max(self, axis: int | None = None, keepdims: bool = False) -> PrivNDArray:
        if axis is None:
            raise DPError("max() with axis=None would collapse the distance axis.")

        if axis < 0:
            axis = self.ndim + axis

        if not (0 <= axis < self.ndim):
            raise ValueError(f"axis {axis} is out of bounds for array of dimension {self.ndim}.")

        if axis == self._privacy_axis:
            raise DPError("max() along the distance axis is not allowed.")

        if keepdims:
            new_privacy_axis = self._privacy_axis
        else:
            if axis < self._privacy_axis:
                new_privacy_axis = self._privacy_axis - 1
            else:
                new_privacy_axis = self._privacy_axis

        result = self._value.max(axis=axis, keepdims=keepdims)

        return PrivNDArray(value          = result,
                           distance       = self._distance,
                           privacy_axis   = new_privacy_axis,
                           domain         = NDArrayDomain(value_range=self._domain.value_range),
                           parents        = [self],
                           keep_alignment = True)

    @egrpc.method
    def min(self, axis: int | None = None, keepdims: bool = False) -> PrivNDArray:
        if axis is None:
            raise DPError("min() with axis=None would collapse the distance axis.")

        if axis < 0:
            axis = self.ndim + axis

        if not (0 <= axis < self.ndim):
            raise ValueError(f"axis {axis} is out of bounds for array of dimension {self.ndim}.")

        if axis == self._privacy_axis:
            raise DPError("min() along the distance axis is not allowed.")

        if keepdims:
            new_privacy_axis = self._privacy_axis
        else:
            if axis < self._privacy_axis:
                new_privacy_axis = self._privacy_axis - 1
            else:
                new_privacy_axis = self._privacy_axis

        result = self._value.min(axis=axis, keepdims=keepdims)

        return PrivNDArray(value          = result,
                           distance       = self._distance,
                           privacy_axis   = new_privacy_axis,
                           domain         = NDArrayDomain(value_range=self._domain.value_range),
                           parents        = [self],
                           keep_alignment = True)

    @egrpc.method
    def copy(self) -> PrivNDArray:
        return PrivNDArray(value          = self._value.copy(),
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = self.domain,
                           parents        = [self],
                           keep_alignment = True)

    @egrpc.multimethod
    def __matmul__(self, other: _npt.NDArray[Any]) -> PrivNDArray:
        if self.ndim != 2 or other.ndim != 2:
            raise NotImplementedError("matmul currently supports 2D operands only.")

        if self._privacy_axis != 0:
            raise DPError("Matmul would contract the distance axis.")

        if self.shape[1] != other.shape[0]:
            raise ValueError("Shape mismatch for matrix multiplication.")

        return PrivNDArray(value          = self._value @ other,
                           distance       = self._distance,
                           domain         = NDArrayDomain(),
                           parents        = [self],
                           privacy_axis   = self._privacy_axis,
                           keep_alignment = True)

    @__matmul__.register
    def _(self, other: PrivNDArray) -> SensitiveNDArray:
        if self.ndim != 2 or other.ndim != 2:
            raise NotImplementedError("matmul currently supports 2D operands only.")

        if self._privacy_axis != 1:
            raise DPError("Left operand privacy_axis must be 1 (the contracting axis).")
        if other._privacy_axis != 0:
            raise DPError("Right operand privacy_axis must be 0 (the contracting axis).")

        if self.alignment_signature != other.alignment_signature:
            raise DPError("alignment_signature mismatch in PrivNDArray matmul.")

        if self.domain.norm_bound is None:
            raise DPError("Left operand norm_bound is not set.")
        if other.domain.norm_bound is None:
            raise DPError("Right operand norm_bound is not set.")

        new_distance = self._distance * self.domain.norm_bound * other.domain.norm_bound
        return SensitiveNDArray(value     = self._value @ other._value,
                                distance  = new_distance,
                                norm_type = "l2",
                                parents   = [self, other])

    @egrpc.method
    def __rmatmul__(self, other: _npt.NDArray[Any]) -> PrivNDArray: # type: ignore[misc]
        if self.ndim != 2 or other.ndim != 2:
            raise NotImplementedError("matmul currently supports 2D operands only.")

        if self._privacy_axis != 1:
            raise DPError("Matmul would contract the distance axis.")

        if self.shape[0] != other.shape[1]:
            raise ValueError("Shape mismatch for matrix multiplication.")

        return PrivNDArray(value          = other @ self._value,
                           distance       = self._distance,
                           domain         = NDArrayDomain(),
                           parents        = [self],
                           privacy_axis   = self._privacy_axis,
                           keep_alignment = True)

    @egrpc.method
    def __neg__(self) -> PrivNDArray:
        new_vr = _negate_value_range(self._domain.value_range)
        return PrivNDArray(value          = -self._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    def _check_axis_aligned(self, other: PrivNDArray) -> None:
        if self.alignment_signature != other.alignment_signature:
            raise DPError("alignment_signature mismatch in PrivNDArray operation.")
        check_broadcast_privacy_axis(self.shape, other.shape)

    @egrpc.multimethod
    def __add__(self, other: realnum) -> PrivNDArray:
        new_vr = _shift_value_range(self._domain.value_range, float(other))
        return PrivNDArray(value          = self._value + other,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    @__add__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        self._check_axis_aligned(other)
        new_vr = _add_value_ranges(self._domain.value_range, other._domain.value_range)
        return PrivNDArray(value          = self._value + other._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self, other],
                           keep_alignment = True)

    @egrpc.method
    def __radd__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        new_vr = _shift_value_range(self._domain.value_range, float(other))
        return PrivNDArray(value          = other + self._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    @egrpc.multimethod
    def __sub__(self, other: realnum) -> PrivNDArray:
        new_vr = _shift_value_range(self._domain.value_range, -float(other))
        return PrivNDArray(value          = self._value - other,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    @__sub__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        self._check_axis_aligned(other)
        new_vr = _sub_value_ranges(self._domain.value_range, other._domain.value_range)
        return PrivNDArray(value          = self._value - other._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self, other],
                           keep_alignment = True)

    @egrpc.method
    def __rsub__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        new_vr = _rshift_value_range(float(other), self._domain.value_range)
        return PrivNDArray(value          = other - self._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    @egrpc.multimethod
    def __mul__(self, other: realnum) -> PrivNDArray:
        new_vr = _scale_value_range(self._domain.value_range, float(other))
        return PrivNDArray(value          = self._value * other,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    @__mul__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        self._check_axis_aligned(other)
        new_vr = _mul_value_ranges(self._domain.value_range, other._domain.value_range)
        return PrivNDArray(value          = self._value * other._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self, other],
                           keep_alignment = True)

    @egrpc.method
    def __rmul__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        new_vr = _scale_value_range(self._domain.value_range, float(other))
        return PrivNDArray(value          = other * self._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    @egrpc.multimethod
    def __truediv__(self, other: realnum) -> PrivNDArray:
        if other == 0:
            raise ZeroDivisionError("division by zero")
        new_vr = _scale_value_range(self._domain.value_range, 1.0 / float(other))
        return PrivNDArray(value          = self._value / other,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    @__truediv__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        self._check_axis_aligned(other)
        new_vr = _div_value_ranges(self._domain.value_range, other._domain.value_range)
        return PrivNDArray(value          = self._value / other._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self, other],
                           keep_alignment = True)

    @egrpc.method
    def __rtruediv__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        new_vr = _rdiv_value_range(float(other), self._domain.value_range)
        return PrivNDArray(value          = other / self._value,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=new_vr),
                           parents        = [self],
                           keep_alignment = True)

    def _compare(self, other: realnum | PrivNDArray, op: str) -> PrivNDArray:
        other_value: _npt.NDArray[Any] | realnum
        if isinstance(other, PrivNDArray):
            self._check_axis_aligned(other)
            other_value = other._value
            parents = [self, other]
        else:
            other_value = other
            parents = [self]

        if op == "eq":
            result = (self._value == other_value).astype(float)
        elif op == "ne":
            result = (self._value != other_value).astype(float)
        elif op == "lt":
            result = (self._value < other_value).astype(float)
        elif op == "le":
            result = (self._value <= other_value).astype(float)
        elif op == "gt":
            result = (self._value > other_value).astype(float)
        elif op == "ge":
            result = (self._value >= other_value).astype(float)
        else:
            raise RuntimeError

        return PrivNDArray(value          = result,
                           distance       = self._distance,
                           privacy_axis   = self._privacy_axis,
                           domain         = NDArrayDomain(value_range=(0.0, 1.0)), # TODO: support bool domain
                           parents        = parents,
                           keep_alignment = True)

    @egrpc.multimethod
    def __eq__(self, other: realnum) -> PrivNDArray:
        return self._compare(other, "eq")

    @__eq__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        return self._compare(other, "eq")

    @egrpc.multimethod
    def __ne__(self, other: realnum) -> PrivNDArray:
        return self._compare(other, "ne")

    @__ne__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        return self._compare(other, "ne")

    @egrpc.multimethod
    def __lt__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        return self._compare(other, "lt")

    @__lt__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        return self._compare(other, "lt")

    @egrpc.multimethod
    def __le__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        return self._compare(other, "le")

    @__le__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        return self._compare(other, "le")

    @egrpc.multimethod
    def __gt__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        return self._compare(other, "gt")

    @__gt__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        return self._compare(other, "gt")

    @egrpc.multimethod
    def __ge__(self, other: realnum) -> PrivNDArray:  # type: ignore[misc]
        return self._compare(other, "ge")

    @__ge__.register
    def _(self, other: PrivNDArray) -> PrivNDArray:
        return self._compare(other, "ge")

    def _normalize_index(self, index: NDIndex) -> tuple[int | slice | None, ...]:
        if not isinstance(index, tuple):
            index = (index,)

        ellipsis_count = sum(1 for i in index if i is ...)
        if ellipsis_count > 1:
            raise IndexError("an index can only have a single ellipsis ('...')")

        if ellipsis_count == 0:
            return index  # type: ignore

        ellipsis_pos = index.index(...)
        n_explicit = len(index) - 1  # excluding ellipsis
        n_missing = self.ndim - n_explicit
        if n_missing < 0:
            raise IndexError("too many indices for array")

        expanded = (
            index[:ellipsis_pos] +
            (slice(None),) * n_missing +
            index[ellipsis_pos + 1:]
        )
        return expanded  # type: ignore[return-value]

    def _validate_index_at_privacy_axis(self, index: tuple[int | slice | None, ...]) -> None:
        axis_indices = [idx for idx in index if idx is not None]
        if self._privacy_axis < len(axis_indices):
            idx = axis_indices[self._privacy_axis]
            if not (isinstance(idx, slice) and idx == slice(None)):
                raise DPError("Cannot index privacy_axis with anything other than ':'")

    def _compute_new_privacy_axis(self, index: tuple[int | slice | None, ...]) -> int:
        offset = 0
        orig_axis = 0
        for idx in index:
            if orig_axis > self._privacy_axis:
                break
            if idx is None:
                offset += 1
            else:
                if orig_axis < self._privacy_axis and isinstance(idx, int):
                    offset -= 1
                orig_axis += 1
        return self._privacy_axis + offset

    @egrpc.method
    def __getitem__(self, index: NDIndex) -> PrivNDArray:
        normalized = self._normalize_index(index)
        self._validate_index_at_privacy_axis(normalized)
        new_privacy_axis = self._compute_new_privacy_axis(normalized)

        return PrivNDArray(value          = self._value[index],
                           distance       = self._distance,
                           privacy_axis   = new_privacy_axis,
                           domain         = NDArrayDomain(value_range=self.domain.value_range),
                           parents        = [self],
                           keep_alignment = True)

@egrpc.remoteclass
class SensitiveNDArray(Prisoner[_npt.NDArray[_np.floating[Any]]]):
    __array_priority__ = 1000  # To prevent NumPy from treating SensitiveNDArray as a scalar operand
    _norm_type: str

    def __init__(self,
                 value      : Any,
                 distance   : RealExpr,
                 norm_type  : str                     = "l1",
                 *,
                 parents    : Sequence[Prisoner[Any]] = [],
                 accountant : Accountant[Any] | None  = None,
                 ) -> None:
        array = _np.asarray(value, dtype=float)
        self._norm_type = norm_type

        super().__init__(value=array, distance=distance, parents=parents, accountant=accountant)

    @egrpc.property
    def shape(self) -> tuple[int, ...]:
        return self._value.shape # type: ignore

    @egrpc.property
    def ndim(self) -> int:
        return self._value.ndim

    @egrpc.property
    def norm_type(self) -> str:
        return self._norm_type

    @egrpc.method
    def __neg__(self) -> SensitiveNDArray:
        return SensitiveNDArray(value     = -self._value,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    @egrpc.multimethod
    def __add__(self, other: realnum) -> SensitiveNDArray:
        return SensitiveNDArray(value     = self._value + other,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    # FIXME: type mismatch error with numpy>=2.2.0
    @__add__.register
    def _(self, other: _npt.NDArray[Any]) -> SensitiveNDArray:
        if self.shape != other.shape:
            raise ValueError("Shape mismatch in SensitiveNDArray addition.")
        return SensitiveNDArray(value     = self._value + other,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    @__add__.register
    def _(self, other: SensitiveNDArray) -> SensitiveNDArray:
        if self.shape != other.shape:
            raise ValueError("Shape mismatch in SensitiveNDArray addition.")
        if self.norm_type != other.norm_type:
            raise ValueError("Norm type mismatch in SensitiveNDArray addition.")
        return SensitiveNDArray(value     = self._value + other._value,
                                distance  = self._distance + other._distance,
                                norm_type = self.norm_type,
                                parents   = [self, other])

    @egrpc.multimethod
    def __radd__(self, other: realnum) -> SensitiveNDArray:  # type: ignore[misc]
        return SensitiveNDArray(value     = other + self._value,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    @__radd__.register
    def _(self, other: _npt.NDArray[Any]) -> SensitiveNDArray:
        if self.shape != other.shape:
            raise ValueError("Shape mismatch in SensitiveNDArray addition.")
        return SensitiveNDArray(value     = other + self._value,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    @egrpc.multimethod
    def __sub__(self, other: realnum) -> SensitiveNDArray:
        return SensitiveNDArray(value     = self._value - other,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    @__sub__.register
    def _(self, other: _npt.NDArray[Any]) -> SensitiveNDArray:
        if self.shape != other.shape:
            raise ValueError("Shape mismatch in SensitiveNDArray subtraction.")
        return SensitiveNDArray(value     = self._value - other,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    @__sub__.register
    def _(self, other: SensitiveNDArray) -> SensitiveNDArray:
        if self.shape != other.shape:
            raise ValueError("Shape mismatch in SensitiveNDArray subtraction.")
        if self.norm_type != other.norm_type:
            raise ValueError("Norm type mismatch in SensitiveNDArray subtraction.")
        return SensitiveNDArray(value     = self._value - other._value,
                                distance  = self._distance + other._distance,
                                norm_type = self.norm_type,
                                parents   = [self, other])

    @egrpc.multimethod
    def __rsub__(self, other: realnum) -> SensitiveNDArray:  # type: ignore[misc]
        return SensitiveNDArray(value     = other - self._value,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    @__rsub__.register
    def _(self, other: _npt.NDArray[Any]) -> SensitiveNDArray:
        if self.shape != other.shape:
            raise ValueError("Shape mismatch in SensitiveNDArray subtraction.")
        return SensitiveNDArray(value     = other - self._value,
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    # TODO: support Literal type in egrpc
    @egrpc.method
    def flatten(self, order: str = "C") -> SensitiveNDArray:
        return SensitiveNDArray(value     = self._value.flatten(order=order), # type: ignore
                                distance  = self._distance,
                                norm_type = self.norm_type,
                                parents   = [self])

    def reveal(self,
               *,
               eps   : floating | None = None,
               delta : floating | None = None,
               rho   : floating | None = None,
               scale : floating | None = None,
               mech  : str             = "laplace",
               ) -> _npt.NDArray[_np.floating[Any]]:
        if mech == "laplace":
            from ..mechanism import laplace_mechanism
            return laplace_mechanism(self,
                                     eps   = float(eps)   if eps   is not None else None,
                                     scale = float(scale) if scale is not None else None)
        elif mech == "gaussian":
            from ..mechanism import gaussian_mechanism
            return gaussian_mechanism(self,
                                      eps   = float(eps)   if eps   is not None else None,
                                      delta = float(delta) if delta is not None else None,
                                      rho   = float(rho)   if rho   is not None else None,
                                      scale = float(scale) if scale is not None else None)
        else:
            raise ValueError(f"Unknown DP mechanism: '{mech}'")

    @egrpc.multimethod
    def __getitem__(self, index: PrivNDArray) -> PrivNDArray:
        # TODO: support more types of indexing
        if self._distance.expr != 0:
            raise DPError("Indexing with PrivNDArray requires distance=0")

        if index.ndim != 1:
            raise DPError("Index must be 1-D array")

        vr = index.domain.value_range
        if vr is None:
            raise DPError("Index must have a defined value_range")
        lo, hi = vr
        if lo is None or hi is None:
            raise DPError("Index value_range must have both lower and upper bounds")
        if lo < 0 or hi >= self.shape[0]:
            raise DPError(f"Index value_range [{lo}, {hi}] is out of bounds for array of length {self.shape[0]}")

        result = self._value[index._value.astype(int)]

        return PrivNDArray(value          = result,
                           distance       = index._distance,
                           privacy_axis   = index._privacy_axis,
                           domain         = NDArrayDomain(), # TODO: calculate domain
                           parents        = [index],
                           keep_alignment = True)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import overload

import numpy as _np
import numpy.typing as _npt

from .util import realnum, DPError
from .numpy import PrivNDArray, NDArrayDomain
from .alignment import new_alignment_signature
from . import egrpc

@egrpc.function
def clip_norm(arr: PrivNDArray, bound: realnum, ord: int | None = None) -> PrivNDArray:
    if bound <= 0:
        raise ValueError("`bound` must be positive.")

    ord_value = 2 if ord is None else ord
    if ord_value not in (1, 2):
        raise ValueError("`ord` must be 1, 2, or None.")

    # FIXME: support for privacy_axis > 0
    assert arr._privacy_axis == 0

    value_array = _np.asarray(arr._value, dtype=float)

    if value_array.size == 0:
        clipped = value_array
    elif value_array.ndim == 1:
        clipped = _np.clip(value_array, -float(bound), float(bound))
    else:
        nrows = value_array.shape[0]
        flat_rows = value_array.reshape(nrows, -1)
        norms = _np.linalg.norm(flat_rows, ord=ord_value, axis=1, keepdims=True)

        scales = _np.ones_like(norms, dtype=float)
        _np.divide(bound, norms, out=scales, where=norms > bound)

        broadcast_shape = (nrows,) + (1,) * (value_array.ndim - 1)
        clipped = value_array * scales.reshape(broadcast_shape)

    norm_type = "l1" if ord_value == 1 else "l2"
    new_domain = NDArrayDomain(norm_type=norm_type, norm_bound=float(bound))
    return PrivNDArray(value          = clipped,
                       distance       = arr._distance,
                       privacy_axis   = arr._privacy_axis,
                       domain         = new_domain,
                       parents        = [arr],
                       keep_alignment = True)

@egrpc.function
def normalize(arr: PrivNDArray, ord: int | None = None) -> PrivNDArray:
    ord_value = 2 if ord is None else ord
    if ord_value not in (1, 2):
        raise ValueError("`ord` must be 1, 2, or None.")

    # FIXME: support for privacy_axis > 0
    assert arr._privacy_axis == 0

    eps = 1e-12
    value_array = _np.asarray(arr._value, dtype=float)

    if value_array.size == 0:
        normalized = value_array
    elif value_array.ndim == 1:
        norm = _np.linalg.norm(value_array, ord=ord_value)
        normalized = value_array / (norm + eps)
    else:
        nrows = value_array.shape[0]
        flat_rows = value_array.reshape(nrows, -1)
        norms = _np.linalg.norm(flat_rows, ord=ord_value, axis=1, keepdims=True)
        broadcast_shape = (nrows,) + (1,) * (value_array.ndim - 1)
        normalized = value_array / (norms.reshape(broadcast_shape) + eps)

    norm_type = "l1" if ord_value == 1 else "l2"
    new_domain = NDArrayDomain(norm_type=norm_type, norm_bound=1.0, value_range=(-1.0, 1.0))
    return PrivNDArray(value          = normalized,
                       distance       = arr._distance,
                       privacy_axis   = arr._privacy_axis,
                       domain         = new_domain,
                       parents        = [arr],
                       keep_alignment = True)

@overload
def sample(array: PrivNDArray, /, *, q: float, method: str = "poisson") -> PrivNDArray: ...
@overload
def sample(*arrays: PrivNDArray, q: float, method: str = "poisson") -> tuple[PrivNDArray, ...]: ...

def sample(*arrays: PrivNDArray, q: float, method: str = "poisson") -> PrivNDArray | tuple[PrivNDArray, ...]:
    result = _sample_impl(arrays, q, method)
    if len(arrays) == 1:
        return result[0]
    return result

@egrpc.function
def _sample_impl(arrays: tuple[PrivNDArray, ...], q: float, method: str) -> tuple[PrivNDArray, ...]:
    if len(arrays) == 0:
        raise ValueError("At least one array is required.")

    if not (0.0 < q <= 1.0):
        raise ValueError("Sampling rate q must be in (0, 1].")

    if method != "poisson":
        raise ValueError(f"Unknown sampling method: '{method}'")

    first = arrays[0]

    if not all(arr.alignment_signature == first.alignment_signature for arr in arrays[1:]):
        raise DPError("All arrays must have the same alignment_signature.")

    # FIXME: support for privacy_axis > 0
    assert all(arr._privacy_axis == 0 for arr in arrays)

    effective_max_distance = float(first._distance.max())
    if effective_max_distance != 1.0:
        raise DPError("Subsampling requires adjacent databases (max_distance=1)")

    n = first._value.shape[0]
    mask: _npt.NDArray[_np.bool_] = _np.random.random(n) < q

    parent_accountant = first.accountant
    child_accountant = parent_accountant.create_subsampling_accountant(q)

    sig = new_alignment_signature()
    results: list[PrivNDArray] = []
    for arr in arrays:
        out = PrivNDArray(value          = arr._value[mask],
                          distance       = arr._distance,
                          privacy_axis   = arr._privacy_axis,
                          domain         = arr.domain,
                          parents        = [arr],
                          accountant     = child_accountant,
                          keep_alignment = False)
        out._alignment_signature = sig
        results.append(out)

    return tuple(results)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any

from .util import Accountant, ParallelAccountant, SubsamplingAccountant
from .. import egrpc

DummyBudgetType = None

@egrpc.remoteclass
class DummyAccountant(Accountant[None]):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs, register=False)

    @staticmethod
    def family_name() -> str:
        return "dummy"

    def propagate(self, next_budget_spent: None, parent: Accountant[Any]) -> None:
        pass

    def compose(self, budget1: None, budget2: None) -> None:
        return None

    def zero(self) -> None:
        return None

    def exceeds(self, budget1: None, budget2: None) -> bool:
        return False

    def assert_budget(self, budget: None) -> None:
        pass

    @classmethod
    def normalize_budget(cls, budget: Any) -> None:
        return None

    @staticmethod
    def parallel_accountant() -> type[DummyParallelAccountant]:
        return DummyParallelAccountant

    @staticmethod
    def subsampling_accountant() -> type[SubsamplingAccountant[None]]:
        raise NotImplementedError("Dummy subsampling accountant is not implemented")

class DummyParallelAccountant(ParallelAccountant[None]):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs, register=False)

    @staticmethod
    def family_name() -> str:
        return "dummy"

    def propagate(self, next_budget_spent: None, parent: Accountant[Any]) -> None:
        pass

    def compose(self, budget1: None, budget2: None) -> None:
        return None

    def zero(self) -> None:
        return None

    def exceeds(self, budget1: None, budget2: None) -> bool:
        return False

    @classmethod
    def normalize_budget(cls, budget: Any) -> None:
        return None

    def assert_budget(self, budget: None) -> None:
        pass



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import privjail as pj
import privjail.pandas as ppd

def main(args):
    df = ppd.read_csv("data/adult_train.csv", "schema/adult.json").dropna()

    n = 1000
    eps = args.eps

    print()
    print("Income stats of younger age groups:")
    print(df.sort_values("age").head(n)["income"].value_counts(sort=False).reveal(eps=eps))

    print()
    print("Income stats of older age groups:")
    print(df.sort_values("age").tail(n)["income"].value_counts(sort=False).reveal(eps=eps))

    print()
    print("Average age by income:")
    for income, df_ in df.groupby("income"):
        mean_age = df_["age"].mean(eps=eps)
        print(f"{income}: {mean_age}")

    print()
    print("Average age and hours-per-week by income:")
    print(df.groupby("income")[["age", "hours-per-week"]].mean(eps=eps))

    print()
    print("Crosstab between gender and income")
    print(ppd.crosstab(df["gender"], df["income"]).reveal(eps=eps))

    print()
    print("Sum of capital-gain and capital-loss of PhDs aged under 40")
    print(df[(df["education"] == "Doctorate") & (df["age"] < 40)][["capital-gain", "capital-loss"]].sum().reveal(eps=eps))

    print()
    print("Consumed Privacy Budget:")
    print(pj.budgets_spent())

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--remote", action="store_true")
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=12345)
    parser.add_argument("--eps", type=float, default=1.0)
    args = parser.parse_args()

    if args.remote:
        pj.connect(args.host, args.port)

    main(args)
